# TODO: Future improvement - RETURN on unknown.  This may make parsing simpler,
# however requires a rewind system.  Re-engineered iterator?

import types

class RewindIter:
	def __init__(this, list):
		this.list = list
		this.pos = 0

	def next(this):
		r = this.list[this.pos]
		this.pos += 1
		return r
	
	def rewind(this):
		this.pos -= 1
	

def get_entry(lines):
	entry = (int(lines.next()), lines.next())
	#print entry
	return entry

class DXF_Object:
	def __repr__(this):
		stri = "DXF_" + this.type
		if this.type == "BLOCK":
			stri = stri + " [" + str(this.POLYLINE) + "]"
		elif this.type == "POLYLINE":
			stri += " ["
			for i in this.VERTEX:
				stri += str(i) + ", "
			stri = stri[:-2] + "]"
		return stri
		
# expects that the iname records are included 
def dxf_obj(type_dict, iname, lines):
	this = DXF_Object()
	this.type = iname
	entry = get_entry(lines)
	assert entry == (0, iname)
	while 1:
		entry = get_entry(lines)
		if (entry[0] == 0):
			try:
				action = type_dict[entry[1]]
			except:
				lines.rewind()
				lines.rewind()
				return this
			if (action == None):
				return this
			else:
				name = entry[1]
				lines.rewind()
				lines.rewind()
				result = action[0](action[1], name, lines)
		else:
			try:
				name = type_dict[entry[0]]
			except:
				lines.rewind()
				lines.rewind()
				return this
			result = entry[1]
		if name == '':
			continue
		exec("this." + name + " = result")	
		
# expects to have the iname records included
def dxf_array(type_dict, iname, lines):
	array = []
	while 1:
		entry = get_entry(lines)
		if entry[0] == 0:
			if entry[1] == "SEQEND" or entry[1] == "ENDSEC":
				return array
			else:
				lines.rewind()
				lines.rewind()
				array.append(dxf_obj(type_dict, iname, lines))

dxf_vertex = {8:'', 62:'', 10:'x', 20:'y', 30:'z', 70:'flags', 71:'m_index',
	72:'n_index', 73:'o_index', 74:'p_index'}
dxf_polyline = {66:'', 62:'colour_number', 8:'', 70:'flags', 71:'m_vertices', 
	72:'n_vertices', 'VERTEX':(dxf_array, dxf_vertex)}
dxf_block = {2:'name', 3:'name', 8:'layer_name', 70:'flags', 10:'bp_x',
	20:'bp_y', 30:'bp_z', 'POLYLINE':(dxf_obj, dxf_polyline), 'ENDBLK':None}
dxf_insert = {8:'', 2:'ref', 10:'x', 20:'y', 30:'z', 41:'x_scale', 
	42:'y_scale', 43:'z_scale', 50:'rotate_angle'}

def dxf_handle_blocks(lines):
	return dxf_array(dxf_block, "BLOCK", lines)
		
def dxf_handle_entities(lines):
	return dxf_array(dxf_insert, "INSERT", lines)

def dxf_to_tree(in_file):
	lines = in_file.readlines()
	lines = map(lambda a: a[:-1], lines)
	lines = RewindIter(lines)
	
	# Header section - ignore for now
	assert get_entry(lines) == (0, "SECTION")
	assert get_entry(lines) == (2, "HEADER")

	while not get_entry(lines) == (0, "ENDSEC"):
		pass

	blocks = None
	entities = None

	while 1:
		next = get_entry(lines)
		if next == (0, "EOF"):
			return (blocks, entities)

		assert next == (0, "SECTION")

		next = get_entry(lines)
		lines.rewind()
		lines.rewind()

		if next == (2, "BLOCKS"):
			blocks = dxf_handle_blocks(lines)
		elif next == (2, "ENTITIES"):
			entities = dxf_handle_entities(lines)


def add(a,b):
	return (a[0]+b[0],a[1]+b[1],a[2]+b[2])

def tree_to_code(blocks, entities, out_file, name):
	block_dict = {}

	for block in blocks:
		block_dict[block.name] = block

	
	out_file.write("from Togra import *\n")
	out_file.write("from TograUtil import *\n")

	out_file.write("def %s():\n" % (name,))
	out_file.write("\tmainContainer = Container()\n")
	for entity in entities:
		if entity.type == "INSERT":
			out_file.write("\tsubContainer = Container()\n")
			block = block_dict[entity.ref]
			offset = (float(entity.x), float(entity.y), float(entity.z))
			scale = (float(entity.x_scale), float(entity.y_scale), 
				float(entity.z_scale))
			r_angle = float(entity.rotate_angle)
			block_base = (float(block.bp_x), float(block.bp_y), 
				float(block.bp_z))
			offset = add(offset, block_base)
			if "POLYLINE" in dir(block):
				pl = block.POLYLINE
				colour = pl.colour_number
				print colour
				n_verts = int(pl.m_vertices)
				n_faces = int(pl.n_vertices)
				verts = pl.VERTEX[:n_verts]
				faces = pl.VERTEX[n_verts:]
				tfaces = []
				qfaces = []
				for face in faces:
					if "p_index" in dir(face):
						qfaces.append(face)
					else:
						tfaces.append(face)
				verts = map(lambda a:
					(float(a.x),float(a.y),float(a.z)), 
					verts)
					
				out_file.write(
			"\tv = map(lambda a: Vector(a[0],a[1],a[2]),\\\n")
				out_file.write("\t" + str(verts) + "\n")
				out_file.write("\t)\n")
				out_file.write(
			"\tvList = VectorList(%d)\n" % (n_verts,))
				out_file.write("\tmap(vList.append, v)\n")
				
				tf = []
				if len(tfaces) > 0:
					map(lambda a: tf.__iadd__(
						[int(a.m_index)-1, 
						int(a.n_index)-1, 
						int(a.o_index)-1]), tfaces)
					out_file.write(
			"\tfList = IntList(%d)\n" % (len(tf)))
					out_file.write("\tmap(fList.append,\n")
					out_file.write("\t" + str(tf) + "\n")
					out_file.write("\t)\n")
					out_file.write(
			"\tsr = StaticRender(vList,GL_TRIANGLES,faces=fList)\n")
					out_file.write(
			"\tsubContainer.append(sr)\n")

				qf = []
				if len(qfaces) > 0:
					map(lambda a: qf.__iadd__(
						[int(a.m_index)-1, 
						int(a.n_index)-1, 
						int(a.o_index)-1,
						int(a.p_index)-1]), qfaces)
					out_file.write(
			"\tfList = IntList(%d)\n" % (len(qf)))
					out_file.write("\tmap(fList.append,\n")
					out_file.write("\t" + str(qf) + "\n")
					out_file.write("\t)\n")
					out_file.write(
			"\tsr = StaticRender(vList,GL_QUADS,faces=fList)\n")
					out_file.write(
			"\tsubContainer.append(sr)\n")
			
				out_file.write(
			"\tsubContainer.transformContainer.append(\n")
				out_file.write(
			"\t\tTransform(translate=Vector(\n")
				out_file.write("\t\t\t" + str(offset[0]) + 
					", " + str(offset[1]) + ", " + 
					str(offset[2]))
				out_file.write(")))\n")
				out_file.write(
			"\tmainContainer.append(subContainer)\n")

			#print offset, scale, r_angle

	out_file.write("\treturn mainContainer\n")

def dxf_to_code(in_file, out_file, name):
	(blocks, entities) = dxf_to_tree(in_file)
	tree_to_code(blocks, entities, out_file, name)

if __name__ == '__main__':
	import sys
	try:
		if sys.argv[1] == "-":
			in_file = sys.stdin
		else:
			in_file = open(sys.argv[1], 'r')
		if sys.argv[2] == "-":
			out_file = sys.stdout
		else:
			out_file = open(sys.argv[2], 'w')
		name = sys.argv[3]
	except:
		print "usage: dxf_to_file.py <in> <out> <function name>"
		sys.exit(1)

	dxf_to_code(in_file, out_file, name)
