#!/usr/bin/env python

from distutils.core import setup, Extension
import sys
import os

def read_pkgconfig (str):
    pkgcfg = os.popen('pkg-config %s' % (str))
    line = pkgcfg.readline().strip()
    pkgcfg.close()
    return line

from distutils import sysconfig
sysconfig.get_config_vars()["CFLAGS"] = \
	"-fno-strict-aliasing -DNDEBUG -g -Wall -Wstrict-prototypes -DGTK_DISABLE_DEPRECATED"

mode = 0
glutMode = 0

while (len(sys.argv) > 2):
	if sys.argv[-1] == "DEBUG":
		mode = 1
	elif sys.argv[2] == "GLUT":
		glutMode = 1
	sys.argv = sys.argv[:-1]


srcs = ["Container.c", "IntList.c", "InterpolatedPath.c", "Render.c",
    "RenderControl.c", "StaticRender.c", "Togra.c", "TograUtil.c",
    "Transform.c", "Util.c", "Vector.c", "VectorList.c", 
    "Material.c", "Matrix.c", "FloatList.c", "Reflect.c",
    "TransformContainer.c", "LightObject.c", "StaticSurface.c",
    "DynamicSurface.c", "StaticTexture.c", "Messages.c", "GstTexture.c",
    "WobbleBehaviour.c", "InterpolationBehaviour.c", "RippleBehaviour.c",
    "ODE_Transform.c", "ODE_Mixin.c", "list.c", "Mixins.c", "ProtoMixin.c",
    "ProtoRender.c", "Graph_Mixin.c", "GraphObject.c", "Logo.c",
    "CustomGraph_Mixin.c", "partition_3d.c", "gst/gsttograsink.c" ]

if mode:
	srcs.append("fake_gl.c")

gst_cflags = read_pkgconfig ('--cflags gstreamer-0.10 gstreamer-plugins-base-0.10')
gst_ldflags = read_pkgconfig ('--libs gstreamer-0.10 gstreamer-plugins-base-0.10')
gst_ldflags += ' -lgstvideo-0.10'
pygobj_cflags = read_pkgconfig ('--cflags pygobject-2.0')
pygobj_ldflags = read_pkgconfig ('--libs pygobject-2.0')

if not glutMode:
	gtk_cflags = read_pkgconfig('--cflags gtk+-2.0 gtkglext-1.0')
	gtk_ldflags = read_pkgconfig('--libs gtk+-2.0 gtkglext-1.0 pangox')

os.system("touch src/Togra.c")

srcs = map (lambda x : "src/" + x, srcs)

if mode:
	modules = [Extension("Togra", include_dirs=["include"], sources=srcs)]
else:
        c_args = ['-O0', '-ggdb', '-Wall', '-Wno-unused-macros', '-DGTK_DISABLE_DEPRECATED']
        c_args += gst_cflags.split()
        c_args += pygobj_cflags.split()
        ld_args = gst_ldflags.split()
        ld_args += pygobj_ldflags.split()

	libs = ["GL", "GLU", "ode"]
	if glutMode:
		c_args.append("-DUSE_GLUT")
		libs.append("glut")
	else:
		c_args += gtk_cflags.split()
		ld_args += gtk_ldflags.split()

	modules = [Extension("Togra", include_dirs=["include", 
                    "/usr/include/glib-2.0", 
                    "/usr/lib/glib-2.0/include", 
                    "/usr/include", "/usr/include/glib-2.0" , 
                    "/usr/include/libxml2" ], 
                    sources=srcs, 
                    extra_compile_args=c_args,
                    extra_link_args=ld_args,
                    libraries=libs)]

setup(name = "Vector", version = "0.1", ext_modules = modules)
