#!/usr/bin/env python

from distutils.core import setup, Extension
import sys
import os

def read_pkgconfig (str):
    pkgcfg = os.popen('pkg-config %s' % (str))
    line = pkgcfg.readline().strip()
    pkgcfg.close()
    return line

from distutils import sysconfig
sysconfig.get_config_vars()["CFLAGS"] = \
	"-fno-strict-aliasing -DNDEBUG -g -Wall -Wstrict-prototypes -DGTK_DISABLE_DEPRECATED"

mode = 0
glutMode = 0

while (len(sys.argv) > 2):
	if sys.argv[-1] == "DEBUG":
		mode = 1
	elif sys.argv[2] == "GLUT":
		glutMode = 1
	sys.argv = sys.argv[:-1]


impl_srcs = ["PyRenderControl", "PyContainer", "PyTransformContainer", "PyLogo",
	     "PyVector", "PyStaticRender", "PyTransform", "PyVectorList",
	     "PyIntList", "PyInterpolatedPath", "PyLinearInterpolation",
	     "PyBezierInterpolation", "PyRotationInterpolation", 
	     "PyTranslation", "PyColour", "PyMaterial", "PyStaticTexture",
	     "PyFloatList"]
other_srcs = ["list.c", "Util.c", "Messages.c", "TograUtil.c"]

srcs = map(lambda a: "parse/" + a + "_impl.c", impl_srcs)
srcs += map(lambda a: "src_new/C_" + a + ".c", impl_srcs)
srcs += map(lambda a: "src_new/" + a, other_srcs)
srcs.append("parse/Togra.c")

if mode:
	srcs.append("fake_gl.c")

gst_cflags = read_pkgconfig ('--cflags gstreamer-0.8')
gst_ldflags = read_pkgconfig ('--libs gstreamer-0.8')
pygobj_cflags = read_pkgconfig ('--cflags pygobject-2.0')
pygobj_ldflags = read_pkgconfig ('--libs pygobject-2.0')

if not glutMode:
        gtk_cflags = read_pkgconfig('--cflags gtk+-2.0 gtkglext-1.0')
        gtk_ldflags = read_pkgconfig('--libs gtk+-2.0 gtkglext-1.0')

os.system("touch src/Togra.c")

if mode:
	modules = [Extension("Togra", include_dirs=["include"], sources=srcs)]
else:
        c_args = ['-O0', '-ggdb', '-Werror', '-Wall', '-Wno-unused-macros', '-DGTK_DISABLE_DEPRECATED', '-Wno-unused']
        c_args += gst_cflags.split()
        c_args += pygobj_cflags.split()
        ld_args = gst_ldflags.split()
        ld_args += pygobj_ldflags.split()

	libs = ["GL", "GLU", "ode"]
	if glutMode:
		c_args.append("-DUSE_GLUT")
		libs.append("glut")
	else:
		c_args += gtk_cflags.split()
		ld_args += gtk_ldflags.split()

	modules = [Extension("Togra", include_dirs=["parse", "include_new",
                    "/usr/include", 
                    "/usr/include/libxml2" ], 
                    sources=srcs, 
                    extra_compile_args=c_args,
                    extra_link_args=ld_args,
                    libraries=libs)]

setup(name = "Togra", version = "0.2", ext_modules = modules)
