/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

#include "Vector_Privates.h"

/* empty comment */

staticforward PyTypeObject PyDynamicSurfaceType;

static void C_PyDynamicSurface_Render(PyObject *self, int time, int picking);
static int C_PyDynamicSurface_Compile(PyObject *self);

static IRender PyDynamicSurface_IRender = {C_PyDynamicSurface_Render, 
        C_PyDynamicSurface_Compile};

static IC PyDynamicSurface_Interfaces = {&PyDynamicSurface_IRender, NULL};

static inline int PyDynamicSurface_Check(PyObject *v) {

        return v->ob_type == &PyDynamicSurfaceType;

}

static inline PyObject * PyDynamicSurface_Vectors(PyObject *v) {

        return ((PyDynamicSurface *)v)->vectors;

}

static inline PyObject * PyDynamicSurface_Normals(PyObject *v) {

        return ((PyDynamicSurface *)v)->normals;

}

void C_PyDynamicSurface_Quads_GenNormals(PyObject *vectors, 
                PyObject **normals, PyObject **faces) {

        /*
        PyVectorList *v = (PyVectorList *)vectors;
        PyVectorList *n = (PyVectorList *)C_PyVectorList_OfLength(v->length);
        */
        /* unfinished ! */
}

void C_PyDynamicSurface_QuadStrip_GenNormals(PyObject *vectors, 
		PyObject **normals) {
        
        PyVectorList *v = (PyVectorList *)vectors;
        PyVectorList *n = (PyVectorList *)C_PyVectorList_OfLength(v->length);
        int i;

	assert(!"GENNORMALS NOT IMPLEMENTED FOR DYNAMICSURFACES!");
	
        for (i = 0; i + 2 < v->length; i += 2)
        {
                PyVector *v0 
                       = (PyVector *)C_PyVector_Subtract(v->pyVectors[i + 1],
                                v->pyVectors[i]);
                PyVector *v1 
                       = (PyVector *)C_PyVector_Subtract(v->pyVectors[i + 3],
                                v->pyVectors[i + 1]);

                VL_APPEND((PyObject *)n, C_PyVector_Cross((PyObject *)v0, 
                                                (PyObject *)v1));
        }

        *normals = (PyObject *)n;

}

void *C_PyDynamicSurface_New(PyObject *vectors, PyObject *normals, 
                PyObject *textureCoords, int textureMode, int width, 
		int height, int genNormals) {

        PyDynamicSurface *r 
		= PyObject_New(PyDynamicSurface, &PyDynamicSurfaceType);
	int j;
	int *p;
	
        D_ALLOC(r, "DynamicSurface");

        r->interfaces = &PyDynamicSurface_Interfaces;
        r->vectors = vectors;
        r->normals = normals;
        r->width = width;
        r->height = height;
        r->textureCoords = textureCoords;
        r->textureMode = textureMode;

	r->faces = malloc(height * 2 * sizeof(int));
	p = r->faces;

	r->ds_behaviour = NULL;
	
	for (j = 0; j < height; j++)
	{
		*p++ = j;
		*p++ = height + j;
	}

	r->vector_cache = malloc(3 * sizeof(double) * width * height);
	r->normal_cache = malloc(3 * sizeof(double) * width * height);
	
	if (vectors != NULL) {
	        Py_INCREF(vectors);
	}

        if (normals != NULL) {
                Py_INCREF(normals);
	}

        if (textureCoords != NULL)
                Py_INCREF(textureCoords);

	
	return (void *)r;

}

static void C_PyDynamicSurface_FlatSurface(PyObject *self, double x0, 
		double y0, double x1, double y1) {

	PyDynamicSurface *ss = (PyDynamicSurface *)self;
	
	int i, j;
	double wstep = (x1 - x0)/(ss->width - 1);
	double hstep = (y1 - y0)/(ss->height - 1);

	if (ss->vectors != NULL) {
		Py_DECREF(ss->vectors);
	}
		
	if (ss->normals != NULL) {
		Py_DECREF(ss->normals);
	}
	
	ss->vectors = C_PyVectorList_OfLength(ss->width * ss->height);
	ss->normals = C_PyVectorList_OfLength(ss->width * ss->height);
	
	for (i = 0; i < ss->width; i++)
	{
		for (j = 0; j < ss->height; j++)
		{
			VL_APPEND(ss->vectors, 
				C_PyVector_FromValues(
					x0 + i * wstep, y0 + j * hstep, 0));
			VL_APPEND(ss->normals,
				C_PyVector_FromValues(0,0,1));
		}
	}
}

#define FROM0_BEZIER_VAL(b,c,d,t)				\
	(3*b*(1-t)*(1-t)*t + 3*c*(1-t)*t*t + d*t*t*t)
/*
 * 3b(1-t)(1-t)t + 3c(1-t)tt + dttt ->
 * -6b(1-t)t + 3b(1-t)(1-t) - 3ctt + 6c(1-t)t + 3dtt ->
 *  3b(1-t)(1-t) + (6c-6b)(1-t)t + (3d-3c)tt
 */
#define FROM0_DBEZIER_VAL(b,c,d,t)				\
	(3*b*(1-t)*(1-t) + (6*c-6*b)*(1-t)*t + (3*d-3*c)*t*t)

#define CROSSX(x0,y0,z0,x1,y1,z1)				\
	(y0*z1-y1*z0)
#define CROSSY(x0,y0,z0,x1,y1,z1)				\
	(z0*x1-z1*x0)
#define CROSSZ(x0,y0,z0,x1,y1,z1)				\
	(x0*y1-x1*y0)

static void _GenerateSpline(PyVectorList *v, PyVectorList *n, double *cxa, 
		double *cxb, double *tx, double *cya, double *cyb, double *ty, 
		int width, int height)
{
	
	int i, j;
	double xx[width];
	double xy[width];
	double xz[width];
	double yx[height];
	double yy[height];
	double yz[height];
	double dxx[width];
	double dxy[width];
	double dxz[width];
	double dyx[height];
	double dyy[height];
	double dyz[height];

	for (i = 0; i < width; i++)
	{
		double t = i*1.0/(width - 1);
		xx[i] = FROM0_BEZIER_VAL(cxa[0],cxb[0],tx[0],t);
		xy[i] = FROM0_BEZIER_VAL(cxa[1],cxb[1],tx[1],t);
		xz[i] = FROM0_BEZIER_VAL(cxa[2],cxb[2],tx[2],t);
		dxx[i] = FROM0_DBEZIER_VAL(cxa[0],cxb[0],tx[0],t);
		dxy[i] = FROM0_DBEZIER_VAL(cxa[1],cxb[1],tx[1],t);
		dxz[i] = FROM0_DBEZIER_VAL(cxa[2],cxb[2],tx[2],t);
	}

	for (i = 0; i < height; i++)
	{
		double t = i*1.0/(height - 1);
		yx[i] = FROM0_BEZIER_VAL(cya[0],cyb[0],ty[0],t);
		yy[i] = FROM0_BEZIER_VAL(cya[1],cyb[1],ty[1],t);
		yz[i] = FROM0_BEZIER_VAL(cya[2],cyb[2],ty[2],t);
		dyx[i] = FROM0_DBEZIER_VAL(cya[0],cyb[0],ty[0],t);
		dyy[i] = FROM0_DBEZIER_VAL(cya[1],cyb[1],ty[1],t);
		dyz[i] = FROM0_DBEZIER_VAL(cya[2],cyb[2],ty[2],t);
	}
	
	for (i = 0; i < width; i++)
	{
		for (j = 0; j < width; j++)
		{
			VL_APPEND((PyObject *)(v), C_PyVector_FromValues(
				xx[i]+yx[j],
				xy[i]+yy[j],
				xz[i]+yz[j]
			));


				VL_APPEND((PyObject *)(n), C_PyVector_FromValues(
				CROSSX(dxx[i],dxy[i],dxz[i],
				       dyx[j],dyy[j],dyx[j]),
				CROSSY(dxx[i],dxy[i],dxz[i],
				       dyx[j],dyy[j],dyx[j]),
				CROSSZ(dxx[i],dxy[i],dxz[i],
				       dyx[j],dyy[j],dyx[j])
			));
		}
	}


}

PyObject *PyDynamicSurface_FlatSurface(PyObject *self, PyObject *args) {
	
	double x0, y0, x1, y1;

        if (!PyArg_ParseTuple(args, "dddd", &x0, &y0, &x1, &y1))
        {
                return NULL;
        }

        C_PyDynamicSurface_FlatSurface(self, x0, y0, x1, y1);

        Py_INCREF(Py_None);
        return Py_None;

}

static void C_PyDynamicSurface_SplineSurface(PyObject *self, 
		double *cxa, double *cxb, double *tx,
		double *cya, double *cyb, double *ty)
{

	PyDynamicSurface *ss = (PyDynamicSurface *)self;
	
	if (ss->vectors != NULL) {
		Py_DECREF(ss->vectors);
	}

	if (ss->normals != NULL) {
		Py_DECREF(ss->normals);
	}
	
	ss->vectors = C_PyVectorList_OfLength(ss->width * ss->height);
	ss->normals = C_PyVectorList_OfLength(ss->width * ss->height);

	_GenerateSpline((PyVectorList *)(ss->vectors),
			(PyVectorList *)(ss->normals),
			cxa, cxb, tx, cya, cyb, ty, ss->width, ss->height);
	
	
}

PyObject *PyDynamicSurface_SplineSurface(PyObject *self, PyObject *args)
{

	PyObject *cxa, *cxb, *tx, *cya, *cyb, *ty;
	/*
	 * PyObject *ncxa, *ncxb, *ntx, *ncya, *ncyb, *nty;
	 * float period;
	 */
	
	if (!PyArg_ParseTuple(args, "OOOOOO", &cxa, &cxb, &tx, &cya, 
				&cyb, &ty))
		return NULL;

	C_PyDynamicSurface_SplineSurface(self, ((PyVector *)cxa)->vector,
		((PyVector *)cxb)->vector,((PyVector *)tx)->vector,
		((PyVector *)cya)->vector,((PyVector *)cyb)->vector,
		((PyVector *)ty)->vector);

	Py_INCREF(Py_None);
	return Py_None;

}

PyObject *PyDynamicSurface_New(PyObject *self, PyObject *args, PyObject *kwords)
{

        PyObject *vectors = NULL, *normals = NULL;
        PyObject *textureCoords = NULL;
        int width, height;
        int textureMode = 0;
        int gen_normals = 0;
					
        static char *kwlist[] = {"width", "height", "vectors", "normals", 
		"genNormals", "textureCoords", "textureMode", 
		NULL};
        
        if (!PyArg_ParseTupleAndKeywords(args, kwords, "ii|OOiOi",
		kwlist, &width, &height, &vectors, &normals, 
		&gen_normals, &textureCoords, &textureMode)) { 
		return NULL; 
	}

        /* TODO: Implement check for length of normals */
        
        return (PyObject *)C_PyDynamicSurface_New(vectors, normals, 
                        textureCoords, textureMode, width, height, 
                        gen_normals);
}

static void PyDynamicSurface_Free(PyObject *self) {

        D_FREE(self, "DynamicSurface");
        
        /* TODO */
        
}

void C_PyDynamicSurface_SetTextureCoords(PyObject *obj, PyObject *v) {

        PyDynamicSurface *r = (PyDynamicSurface *)obj;
        
        if (r->textureCoords != NULL)
        {
                Py_DECREF(r->textureCoords);
        }
        
        Py_INCREF(v);
        r->textureCoords = v;


}

static int PyDynamicSurface_Setattr(PyObject *obj, char *name, PyObject *v) {
	
	/*PyDynamicSurface *r = (PyDynamicSurface *)obj;*/

	if (strcmp(name, "textureCoords") == 0) {
                C_PyDynamicSurface_SetTextureCoords(obj, v);
                return 0;
        }
	
	return -1;
}

void C_PyDynamicSurface_SetBehaviour(PyObject *this, PyObject *behaviour) {

	PyDynamicSurface *ds = (PyDynamicSurface *)this;
	
	if (ds->ds_behaviour != NULL) {
		Py_DECREF(ds->ds_behaviour);
	}
	
	if (behaviour == Py_None) {
		ds->ds_behaviour = NULL;
		return;
	}

	Py_INCREF(behaviour);
	ds->ds_behaviour = behaviour;

}

PyObject *PyDynamicSurface_SetBehaviour(PyObject *this, PyObject *args) {

	PyObject *behaviour;
	
	if (!PyArg_ParseTuple(args, "O", &behaviour))
		return NULL;

	C_PyDynamicSurface_SetBehaviour(this, behaviour);

	Py_INCREF(Py_None);
	return Py_None;
	
}

static PyMethodDef PyDynamicSurface_methods[] = {
	{"setBehaviour", (PyCFunction)PyDynamicSurface_SetBehaviour,
		METH_VARARGS, "set the dynamic behaviour of this surface"},
	{"flatSurface", (PyCFunction)PyDynamicSurface_FlatSurface,
		METH_VARARGS, "set this surface up as a flat surface in x-y"},
	{"splineSurface", (PyCFunction)PyDynamicSurface_SplineSurface,
		METH_VARARGS, "set this surface up as a spline surface in x-y"},
	{NULL, NULL, 0, NULL}
};


static PyObject *PyDynamicSurface_Getattr(PyObject *obj, char *name) {
	
	PyDynamicSurface *r = (PyDynamicSurface *)obj;

        if (strcmp(name, "textureCoords") == 0) {

                if (r->textureCoords == NULL) {
                        Py_INCREF(Py_None);
                        return Py_None;
                }

                Py_INCREF(r->textureCoords);
                return r->textureCoords;

        }
	
	return Py_FindMethod(PyDynamicSurface_methods, obj, name);
	
}

static int C_PyDynamicSurface_Compile(PyObject *self) {

	return FALSE;
}

static void C_PyDynamicSurface_Render(PyObject *self, int time, int picking) {

        PyDynamicSurface *r = (PyDynamicSurface *)self;
	int i;
	double *vp = ((PyVectorList *)(r->vectors))->values;
	double *np = ((PyVectorList *)(r->normals))->values;
	
	memcpy(r->vector_cache, vp, sizeof(double) * r->width * r->height * 3);
	memcpy(r->normal_cache, np, sizeof(double) * r->width * r->height * 3);

	/*
	for (i = 0; i < r->width; i++)
	{
		int j;
		for (j = 0; j < r->height; j++)
		{
			printf
			(
			 	"%f %f %f\n", 
				r->vector_cache[3 * i * r->height + 3 * j],
				r->vector_cache[3 * i * r->height + 3 * j + 1],
				r->vector_cache[3 * i * r->height + 3 * j + 2]
			);
		}
	}
	*/
	
	if (r->ds_behaviour != NULL)
		DSBEHAVIOUR_RUN(r->ds_behaviour, r->vector_cache, 
			r->normal_cache, r->width, r->height, time);
	
	/*
	r->dynamic_function(r, r->vector_cache, r->normal_cache, 
			r->dynamic_data, time);
	*/

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	
	if (r->textureCoords != NULL)
	{
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	}

	for (i = 0; i < r->width - 1; i++)
	{
		glVertexPointer(3, GL_DOUBLE, 0, 
				r->vector_cache + (3 * i * r->height));
		glNormalPointer(GL_DOUBLE, 0, 
				r->normal_cache + (3 * i * r->height));
		if (r->textureCoords != NULL)
		{
			float *arr;
			arr = ((PyFloatList *)(r->textureCoords))->floats;
			arr += 2 * i * r->height;
			
			glTexCoordPointer(2, GL_FLOAT, 0, arr);
		}
		glDrawElements(GL_QUAD_STRIP, 2*r->height, 
				GL_UNSIGNED_INT,
				r->faces);
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

static PyObject *PyDynamicSurface_Call(PyObject *self, PyObject *args, PyObject *dict) {

        PyObject *p;
        int gl_type = GL_POINTS;


        if (!(PyTuple_Size(args) == 1)) {
                return NULL;
        }

        p = PyTuple_GetItem(args, 0);

        if (p != Py_None) {

                if (!(PyInt_Check(p))) {
                        return NULL;
                }

                gl_type = (int) PyInt_AsLong(p);

        }

        C_PyDynamicSurface_Render((void *)self, 0, 0);

        Py_INCREF(Py_None);
        return Py_None;

}
        
static PySequenceMethods PyDynamicSurface_sequenceMethods = {
        0, /* length */
        0, /* concat */
        0, /* repeat */
        0, /* item */
        0, /* slice */
        0, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyDynamicSurfaceType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyDynamicSurface",
        sizeof(PyDynamicSurface),
        0,
        PyDynamicSurface_Free,
        0, /* print */
        PyDynamicSurface_Getattr, /* getattr */
        PyDynamicSurface_Setattr, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyDynamicSurface_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        &PyDynamicSurface_Call, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
