/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"
staticforward PyTypeObject PyFloatListType;

static int C_PyFloatList_Item(PyObject *self, int index, float *store);
static int PyFloatList_Length(PyObject *self);

static inline int PyFloatList_Check(PyObject *v) {

        return v->ob_type == &PyFloatListType;

}

static int PyFloatList_Length(PyObject *v) {

        return ((PyFloatList *)v)->length;

}

static inline int PyFloatList_Alloced(PyObject *v) {

        return ((PyFloatList *)v)->alloced;

}

PyObject *C_PyFloatList_New(void) {

        PyFloatList *l = PyObject_New(PyFloatList, &PyFloatListType);

        D_ALLOC(l, "FloatList");

        l->alloced = l->length = 0;
        l->interfaces = NULL;
        l->floats = NULL;

        return (PyObject *)l;
        
}

PyObject *C_PyFloatList_OfLength(int length) {

        PyFloatList *l = (PyFloatList *)C_PyFloatList_New();
        l->alloced = length;
        
        if (l->alloced > 0) {
                l->floats = malloc(sizeof(float) * l->alloced);
        }

        return (PyObject *)l;

}
        
PyObject *PyFloatList_New(PyObject *self, PyObject *args) {

        PyObject *l;
        int size;

        if (PyTuple_Size(args) == 1) {

                l = PyTuple_GetItem(args, 0);

                if (!PyInt_Check(l)) {
                        return NULL;
                }

                size = (int)PyInt_AsLong(l);

                return (PyObject *)C_PyFloatList_OfLength(size);

        } else if (PyTuple_Size(args) == 0) {

                return (PyObject *)C_PyFloatList_New();

        }

        return NULL;
        
}

static void C_PyFloatList_Free(PyObject *obj) {

        PyFloatList *l = (PyFloatList *)obj;

        D_FREE(l, "FloatList");
        
        if (l->alloced > 0) {
                free(l->floats);
        }

        PyObject_Del((PyObject *)obj);
}

static void PyFloatList_Free(PyObject *self) {

        C_PyFloatList_Free((void *)self);
        
}

static int C_PyFloatList_Item(PyObject *self, int index, float *store) {
	
	PyFloatList *l = (PyFloatList *)self;

	if (index > l->length) {
		
		return -1;
		
	}
	
	*store = l->floats[index];
	return 0;
	
}

static PyObject *PyFloatList_Item(PyObject *self, int index) {
	
	float val;
	
	if (C_PyFloatList_Item(self, index, &val) == -1) {
		return NULL;
	}
	
	return PyFloat_FromDouble(val);
	
}
	

void C_PyFloatList_Append(PyObject *obj, float val) {

        PyFloatList *l = (PyFloatList *)obj;

        if (l->alloced == l->length) {
                
                if (l->alloced == 0) {
                        l->alloced++;
                } else {
                        l->alloced *= 2;
                }

                l->floats = realloc(l->floats, l->alloced * sizeof(float));

        }

        
        l->floats[l->length++] = val;


}

static PyObject *PyFloatList_Append(PyObject *self, PyObject *args) {

        float val;
        
        if (!PyArg_ParseTuple(args, "f", &val)) {
                return NULL;
        }

        C_PyFloatList_Append(self, val);

        Py_INCREF(Py_None);
        return Py_None;

}

static PyMethodDef PyFloatList_methods[] = {
        {"append", (PyCFunction)PyFloatList_Append, METH_VARARGS,
                "Appends an integer to the int list"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyFloatList_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyFloatList_methods, self, name);

}
        
static PySequenceMethods PyFloatList_sequenceMethods = {
        PyFloatList_Length, /* length */
        0, /* concat */
        0, /* repeat */
        PyFloatList_Item, /* item */
        0, /* slice */
        0, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyFloatListType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyFloatList",
        sizeof(PyFloatList),
        0,
        PyFloatList_Free,
        0, /* print */
        PyFloatList_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyFloatList_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
