/*
 * Copyright 2003, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

staticforward PyTypeObject PyInterpolationBehaviourType;

void C_PyInterpolationBehaviour_Run(PyObject *self, double *v, double *n, int width,
		int height, int time);
static IDSBehaviour PyInterpolationBehaviour_IDSBehaviour 
	= {C_PyInterpolationBehaviour_Run};

static IC PyInterpolationBehaviour_Interfaces = {NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, &PyInterpolationBehaviour_IDSBehaviour};

void C_PyInterpolationBehaviour_Run(PyObject *self, double *v, double *n, int width,
		int height, int time) {

	int i;
	
	PyInterpolationBehaviour *d = (PyInterpolationBehaviour *)self;
	PyDynamicSurface *ds = (PyDynamicSurface *)(d->to_surface);
	double *ev = ((PyVectorList *)(ds->vectors))->values;
	double *en = ((PyVectorList *)(ds->normals))->values;
	double s, e;

	time %= d->period;
	if (time > d->period / 2)
		time = d->period - time;

	s = -cos(3.142 * 2 * time/(float)(d->period))/2 + 0.5;
	e = 1 - s;

	for (i = 0; i < width * height * 3; i++)
	{
		v[i] = s*v[i] + e*ev[i];
		n[i] = s*n[i] + e*en[i];
	}
	
}

PyObject *C_PyInterpolationBehaviour_New(PyObject *to_surface, int period) {
	
	PyInterpolationBehaviour *d = PyObject_New(PyInterpolationBehaviour,
			&PyInterpolationBehaviourType);


	d->interfaces = &PyInterpolationBehaviour_Interfaces;
	
	Py_INCREF(to_surface);
	d->to_surface = to_surface;
	d->period = period;
	
	return (PyObject *)d;
	
}

PyObject *PyInterpolationBehaviour_New(PyObject *self, PyObject *args) {

	PyObject *dyn;
	int time;
	
	if (!PyArg_ParseTuple(args, "Oi", &dyn, &time)) {
		return NULL;
	}

	return C_PyInterpolationBehaviour_New(dyn, time);
	
}


/*
static PyObject *PyInterpolationBehaviour_Setattr(PyObject *obj, char *name,
		PyObject *val) {

	PyInterpolationBehaviour *data = (PyInterpolationBehaviour *)obj;
	
	if (strcmp(name, "xPeriodPeriod") == 0)
	{
		data->xperper = PyFloat_AsDouble(val);
	}
	else if (strcmp(name, "yPeriodPeriod") == 0)
	{
		data->yperper = PyFloat_AsDouble(val);
	}
	else if (strcmp(name, "xPeriodMagnitude") == 0)
	{
		data->xpermag = PyFloat_AsDouble(val);
	}
	else if (strcmp(name, "yPeriodMagnitude") == 0)
	{
		data->ypermag = PyFloat_AsDouble(val);
	}
	else
	{
		return -1;
	}

	return 0;
}
*/

static void PyInterpolationBehaviour_Free(PyObject *this) {

}

static PyTypeObject PyInterpolationBehaviourType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyInterpolationBehaviour",
        sizeof(PyInterpolationBehaviour),
        0,
        PyInterpolationBehaviour_Free,
        0, /* print */
        0, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
