#include "TograIncludes.h"

staticforward PyTypeObject PyLightObjectType;
static void C_PyLightObject_Render(PyObject *self, int time, int picking);
static int C_PyLightObject_Compile(PyObject *self) {return 0;}

static IRender PyLightObject_IRender = {C_PyLightObject_Render,
                C_PyLightObject_Compile};
static IC PyLightObject_Interfaces = {&PyLightObject_IRender, NULL};

PyObject *C_PyLightObject_New(void) {
        
        PyLightObject *l = PyObject_New(PyLightObject, &PyLightObjectType);

        D_ALLOC(l, "LightObject");
        
        l->interfaces = &PyLightObject_Interfaces;
        l->light = C_PyRenderControl_NewLight();

#if DEBUG_VERBOSE
	printf("light is %d\n", l->light);
#endif

        return (PyObject *)l;

}

int PyContainer_Check(PyObject *obj);
void C_PyTransformContainer_Append(PyObject *tc, PyObject *obj);

int _add_rc(PyTransformContainer *tc, PyContainer *c, PyObject *this)
{

        int i;

        int parent = 0;

        for (i = 0; i < c->length; i++)
        {
                if (PyContainer_Check(c->children[i])) {
                        if (_add_rc(tc, (PyContainer *)(c->children[i]), 
						this)) {
                                parent = 1;
                                break;
                        }
                } else if (c->children[i] == this) {
                        parent = 1;
                        break;
                }
        
        }

        if (parent == 0) {
                return 0;
        }
        
        if (c->transform != NULL)
                C_PyTransformContainer_Append((PyObject *)tc, c->transform);
        
        return 1;
}
        
extern PyObject *this_ptr;
void C_PyRenderControl_SetLightPosition(int light, float *pos);
void C_PyRenderControl_SetLightDiffuse(int light, float *diffuse);

PyObject *PyLightObject_New(PyObject *this, PyObject *args, PyObject *kwords) {

        PyObject *location = NULL, *diffuse = NULL;
        int transformWithTree = 1;
        float pos[4] = {0,0,0,1};
        float dif[4] = {1,1,1,1};
        int i;

        static char *kwlist[] = {"location", "diffuse",  
                "transformWithTree", NULL};

        PyLightObject *l = (PyLightObject *)C_PyLightObject_New();
        
        if (!PyArg_ParseTupleAndKeywords(args, kwords, "|OOi", kwlist,
                                &location, &diffuse, 
                                &transformWithTree))
        {
                return NULL;
        }

        if (location != NULL)
        {
                for (i = 0; i < 3; i++)
                        pos[i] = ((PyVector *)location)->vector[i];
                C_PyRenderControl_SetLightPosition(l->light, pos);
        }

        if (diffuse != NULL)
        {
                for (i = 0; i < 4; i++)
                        dif[i] = (float)PyFloat_AsDouble(
                                        PyTuple_GetItem(diffuse, i));
                C_PyRenderControl_SetLightDiffuse(l->light, dif);
        }

        l->transformWithTree = transformWithTree;

        return (PyObject *)l;
        
}

void C_PyRenderControl_SetLightTransform(int light, PyObject *transform);

static void C_PyLightObject_Render(PyObject *self, int time, int picking)
{

        PyLightObject *l = (PyLightObject *)self;
        PyTransformContainer *tc;
        
        if (l->transformWithTree) {
                l->transformWithTree = 0;
                tc = (PyTransformContainer *)C_PyTransformContainer_New();
                PyContainer *c = (PyContainer *)
				(((PyRenderControl *)this_ptr)->renderObject);

                if (_add_rc(tc, c, self)) {
#if DEBUG_VERBOSE
			printf("Setting transform\n");
#endif
                	C_PyRenderControl_SetLightTransform(l->light, 
						(PyObject *)tc);
		}
#if DEBUG_VERBOSE
		else {
			printf("Not setting transform\n");
		}
#endif
        }
}

static void PyLightObject_Free(PyObject *self) {

        D_FREE(self, "LightObject");

}

static PyTypeObject PyLightObjectType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyLightObject",
        sizeof(PyLightObject),
        0,
        PyLightObject_Free,
        0, /* print */
        0, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};

