/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

/* empty comment */

staticforward PyTypeObject PyLogoType;

static void C_PyLogo_Render(PyObject *self, int time, int picking);
static int C_PyLogo_Compile(PyObject *self) {return FALSE;}

static IRender PyLogo_IRender = {C_PyLogo_Render, C_PyLogo_Compile};
static IC PyLogo_Interfaces = {&PyLogo_IRender, NULL};

#define VERTICAL_BYPARAM	0
#define TOP			1
#define VERTICAL_CENTER		2
#define BOTTOM			3
#define HORIZONTAL_BYPARAM	0
#define LEFT			4
#define HORIZONTAL_CENTER	8
#define RIGHT			12

PyObject *C_PyStaticTexture_New(int *bits, int width, int height, int mode);
PyObject *C_PyLogo_New(int *l, int x, int y, int width, int height, 
		int mode) {

	int w = 2, h = 2;
	int *lnew;
	int i;
	
        PyLogo *r = PyObject_New(PyLogo, &PyLogoType);

	while (w < width)
		w *= 2;
	
	while (h < height)
		h *= 2;
	
	lnew = malloc(w* h * 4);
	
	for (i = 0; i < height; i++)
		memcpy(lnew + i * w, l + i * width, width * 4);
	
	r->logo = C_PyStaticTexture_New(lnew, w, h, mode);
	
	free(lnew);
	
	r->interfaces = &PyLogo_Interfaces;

	r->width = width;
	r->height = height;
	r->x = x;
	r->xfrac = ((float)width) / w;
	r->y = y;
	r->yfrac = ((float)height) / h;
        r->mode = mode;
	r->position = 0;

        return (PyObject *)r;
        
}

PyObject *PyLogo_New(PyObject *self, PyObject *args) {

	PyObject *logo;
	int width, height, mode, x, y;
					
        if (!PyArg_ParseTuple(args, "Oiiiii", &logo, &x, &y, &width, &height, 
				&mode)) {
                return NULL;
        }

        return (PyObject *)C_PyLogo_New((int *)PyString_AsString(logo),
			x, y, width, height, mode);

}

PyObject *PyLogo_SetPositioning(PyObject *self, PyObject *args) {

	int pos;
	PyLogo *r = (PyLogo *)self;

	if (!PyArg_ParseTuple(args, "i", &pos))
		return NULL;
	
	r->position = pos;

	Py_INCREF(Py_None);
	return Py_None;
	
}

static PyMethodDef PyLogo_methods[] = {
	{"setPositioning", (PyCFunction)PyLogo_SetPositioning, METH_VARARGS,
		"Sets the position parameters of the logo"},
	{NULL, NULL, 0, NULL}
};

static PyObject *PyLogo_GetAttr(PyObject *self, char *name) {

	return Py_FindMethod(PyLogo_methods, self, name);

}

static void PyLogo_Free(PyObject *self) {
        
}

int getScreenWidth(void);
int getScreenHeight(void);

#define GET_YPOS(x)	(x & 0x3)
#define GET_XPOS(x)	(x & 0xC)

static void C_PyLogo_Render(PyObject *self, int time, int picking) {

        PyLogo *r = (PyLogo *)self;

	int width = getScreenWidth();
	int height = getScreenHeight();
	int xpos, ypos;
	float col[] = {1.0,1.0,1.0,1.0};

	if (GET_YPOS(r->position) == TOP)
		ypos = 0;
	else if (GET_YPOS(r->position) == VERTICAL_CENTER)
		ypos = (height - r->height) / 2;
	else if (GET_YPOS(r->position) == BOTTOM)
		ypos = height - r->height;
	else
		ypos = r->y;

	if (GET_XPOS(r->position) == LEFT)
		xpos = 0;
	else if (GET_XPOS(r->position) == HORIZONTAL_CENTER)
		xpos = (width - r->width) / 2;
	else if (GET_XPOS(r->position) == RIGHT)
		xpos = width - r->width;
	else
		xpos = r->x;
	
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	
	gluOrtho2D(0.0, width, height, 0.0);

	/*
	glRasterPos2i(xpos, ypos);
	glPixelZoom(1.0, -1.0);
	glDrawPixels(r->width, r->height, r->mode, GL_UNSIGNED_BYTE, logo_bits);
	*/
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, col);
        if (!picking)
	    TEXTURE_SET(r->logo, time);
	
	glBegin(GL_QUADS);
	glNormal3f(0,0,1);
	glTexCoord2f(0.0, 0.0);
	glVertex3f(xpos, ypos, 0);
	glTexCoord2f(0.0, r->yfrac);
	glVertex3f(xpos, ypos + r->height, 0);
	glTexCoord2f(r->xfrac, r->yfrac);
	glVertex3f(xpos + r->width, ypos + r->height, 0);
	glTexCoord2f(r->xfrac, 0.0);
	glVertex3f(xpos + r->width, ypos, 0);
	glEnd();
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

}

static PyTypeObject PyLogoType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyLogo",
        sizeof(PyLogo),
        0,
        PyLogo_Free,
        0, /* print */
        PyLogo_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
