#include "TograIncludes.h"

staticforward PyTypeObject PyMatrixType;

/*
typedef struct {
        void (*generateTransform)(PyObject *self, PyObject *status, int time, 
                        int gen_forward, int gen_reverse, int apply);
        void (*transformByMe)(PyObject *self, PyObject *status, int time);
        void (*invertByMe)(PyObject *self, PyObject *status, int time);
        int (*registerSize)(PyObject *self);
        void (*registerFunc)(PyObject *self, PyObject *status, int delay,
                        PyObject *callback, PyObject *user, PyObject *user2);
        void (*deRegisterFunc)(PyObject *self, PyObject *status);
} ITransform;
*/

static void C_PyMatrix_GenerateTransform(PyObject *self, PyObject *status, 
		int time, int gen_forward, int gen_reverse, int apply);
inline void C_PyMatrix_TransformByMe(PyObject *self);
inline void C_PyMatrix_InvertByMe(PyObject *self);
int C_PyMatrix_RegisterSize(PyObject *self) { return 0; }

static ITransform PyMatrix_ITransform = {C_PyMatrix_GenerateTransform,
        C_PyMatrix_TransformByMe, C_PyMatrix_InvertByMe, 
        C_PyMatrix_RegisterSize, NULL, NULL};

static IC PyMatrix_Interfaces = {NULL, NULL, NULL, NULL, NULL, 
                                        &PyMatrix_ITransform};

static void _C_PyMatrix_InvertTo(double *m, double *i);

PyObject *C_PyMatrix_New(void) {

        PyMatrix *m = (PyMatrix *)PyObject_New(PyMatrix, &PyMatrixType);

        D_ALLOC(m, "Matrix");
        
        memset(m->values, 0, 16 * sizeof(double));
        m->values[0] = m->values[5] = m->values[10] = m->values[15] = 1;
        m->interfaces = &PyMatrix_Interfaces;
        m->hasInverse = 0;
        
        return (PyObject *)m;
}

PyObject *C_PyMatrix_FromMatrix(double *mm)
{
	PyMatrix *m = (PyMatrix *)C_PyMatrix_New();
	memcpy(m->values, mm, 16 * sizeof(double));

	return (PyObject *)m;

}

PyObject *PyMatrix_FromMatrix(PyObject *self, PyObject *args)
{
	PyObject *list = PyTuple_GetItem(args, 0);
	double m[16];
	int i;

	for(i = 0; i < 16; i++)
	{
		m[i] = PyFloat_AsDouble(PyList_GetItem(list, i));
	}

	return C_PyMatrix_FromMatrix(m);

}

void C_PyMatrix_Free(PyObject *self) {

        D_FREE(self, "Matrix");
        
        PyObject_Del(self);
        
}
        
PyObject *C_PyMatrix_FromTranslation(double *translation) {

        PyMatrix *m = (PyMatrix *)C_PyMatrix_New();

        memcpy(&(m->values[12]), translation, 3 * sizeof(double));

        return (PyObject *)m;
}

PyObject *C_PyMatrix_FromRotation(double *rotateAxis, double rotateAmt) {

        /*
         * mathematics from: http://www.makegames.com/3drotation/
         *      (Diana Gruber)
         */

        PyMatrix *m = (PyMatrix *)C_PyMatrix_New();
	float _tv[16];
	int i;
	
	glPushMatrix();
	glLoadIdentity();
	glRotated(rotateAmt, rotateAxis[0], rotateAxis[1], rotateAxis[2]);
	glGetFloatv(GL_MODELVIEW_MATRIX, _tv);
	glPopMatrix();

	for (i = 0; i < 16; i++)
	{
		m->values[i] = (double)_tv[i];
	}
	
	/*
	double c = cos(rotateAmt*M_PI/180);
        double s = sin(rotateAmt*M_PI/180);
        double t = 1 - c;
        double x = rotateAxis[0];
        double y = rotateAxis[1];
        double z = rotateAxis[2];

        m->values[0] = t*x*x + c;
        m->values[1] = t*x*y + s*z;
        m->values[2] = t*x*z - s*y;
        m->values[4] = t*x*y - s*z;
        m->values[5] = t*y*y + c;
        m->values[6] = t*y*z + s*x;
        m->values[8] = t*x*z + s*y;
        m->values[9] = t*y*z - s*x;
        m->values[10] = t*z*z + c;
	*/
        return (PyObject *)m;
        
}

PyObject *PyMatrix_NewTranslation(PyObject *self, PyObject *args) {

        PyObject *translation;

        if (!PyArg_ParseTuple(args, "O", &translation)) {
                return NULL;
        }
        
        return C_PyMatrix_FromTranslation(((PyVector *)translation)->vector);
}

PyObject *PyMatrix_NewRotation(PyObject *self, PyObject *args) {

        PyObject *rotateAxis;
        double rotateAmount;

        if (!PyArg_ParseTuple(args, "Od", &rotateAxis, &rotateAmount)) {
                return NULL;
        }
        
        return C_PyMatrix_FromRotation(((PyVector *)rotateAxis)->vector, 
                                                        rotateAmount);
        
}

inline void C_PyMatrix_TransformByMe(PyObject *self) {

        glMultMatrixd(((PyMatrix *)self)->values);

}

inline void C_PyMatrix_InvertByMe(PyObject *self) {

        glMultMatrixd(((PyMatrix *)self)->inverse);

}

void C_PyMatrix_GenerateTransform(PyObject *self, PyObject *status, int time,
                int gen_forward, int gen_reverse, int apply) {

        PyMatrix *m = (PyMatrix *)self;
        
        if (gen_reverse)
        {
                _C_PyMatrix_InvertTo(m->values, m->inverse);
        }

        if (apply == 1)
                C_PyMatrix_TransformByMe(self);
        else if (apply == -1)
                C_PyMatrix_InvertByMe(self);

}
        

PyObject *C_PyMatrix_FromScale(double *scale) {

        PyMatrix *m = (PyMatrix *)C_PyMatrix_New();

        m->values[0] = scale[0];
        m->values[5] = scale[1];
        m->values[10] = scale[2];

        return (PyObject *)m;

}

PyObject *PyMatrix_NewScale(PyObject *self, PyObject *args) {

        PyObject *scale;

        if (!PyArg_ParseTuple(args, "O", &scale)) {
                return NULL;
        }
        
        return C_PyMatrix_FromScale(((PyVector *)scale)->vector);
        
}

/*
 *  0  4  8  12
 *  1  5  9  13
 *  2  6  10 14
 *  3  7  11 15
 *
 * 0 <- a0*b0 + a4*b1 + a8*b2 + a12*b3
 */
PyObject *C_PyMatrix_Multiply(PyObject *a, PyObject *b)
{
        PyMatrix *ma, *mb;
        PyMatrix *result;
        int i, j, k;
        
        ma = (PyMatrix *)a;
        mb = (PyMatrix *)b;

        result = (PyMatrix *)C_PyMatrix_New();
       
        for (i = 0; i < 4; i++)
        {
                for (j = 0; j < 4; j++)
                {
                        //printf("Calc: %d\n", i*4+j);
                
                        result->values[i*4 + j] = 0;
                        for (k = 0; k < 4; k++)
                        {
                        //printf("%d * %d\n", k*4+j, i*4+k);
                                result->values[i*4+j] += 
                                        ma->values[k*4 + j]
                                        *
                                        mb->values[i*4 + k];
                        }
                }
        }
        
        return (PyObject *)result;

}

PyObject *C_PyMatrix_MultiplyVector(PyObject *m, PyObject *v) {

	PyMatrix *mm;
	PyVector *vv;
	PyVector *result;
	int i, j;

	mm = (PyMatrix *)m;
	vv = (PyVector *)v;

	result = (PyVector *)C_PyVector_New();

	for (i = 0; i < 3; i++) {

		result->vector[i] = 0;

		for (j = 0; j < 3; j++) {
		
			result->vector[i] += mm->values[j*4 + i] *
				vv->vector[j];
		}
		result->vector[i] += mm->values[12 + i];
	}

	return (PyObject *)result;
	
}

PyObject *PyMatrix_Multiply(PyObject *self, PyObject *args) {

        PyObject *a;
        
        if (!PyArg_ParseTuple(args, "O", &a)) {
                return NULL;
        }
        
        return C_PyMatrix_Multiply(self, a);
        
}

PyObject *PyMatrix_MultiplyVector(PyObject *self, PyObject *args) {

	PyObject *a;

	if (!PyArg_ParseTuple(args, "O", &a)) {
		return NULL;
	}

	return C_PyMatrix_MultiplyVector(self, a);
}

void C_PyMatrix_Print(PyObject *matrix, FILE *fp, int flags) {

        double *m = ((PyMatrix *)matrix)->values;
        
        fprintf(fp, "/ %4.2f %4.2f %4.2f %4.2f \\\n", 
                                                m[0], m[4], m[8], m[12]);
        fprintf(fp, "| %4.2f %4.2f %4.2f %4.2f |\n",  
                                                m[1], m[5], m[9], m[13]);
        fprintf(fp, "| %4.2f %4.2f %4.2f %4.2f |\n",  
                                                m[2], m[6], m[10], m[14]);
        fprintf(fp, "\\ %4.2f %4.2f %4.2f %4.2f /\n",  
                                                m[3], m[7], m[11], m[15]);

}

inline double det2(double *m, int r1, int c1, int r2, int c2)
{

        double r = m[4*c1+r1] * m[4*c2+r2] - m[4*c1+r2] * m[4*c2+r1];
        //printf("det2 %d %d %d %d %f\n", r1, r2, c1, c2, r);
        return r;

}

inline double det3(double *m, int r1, int r2, int r3, int c1, int c2, int c3)
{
        double r = m[4*c1+r1] * det2(m, r2, c2, r3, c3) -
               m[4*c2+r1] * det2(m, r2, c1, r3, c3) +
               m[4*c3+r1] * det2(m, r2, c1, r3, c2);
        //printf("det3 %d %d %d %d %d %d %f\n", r1, r2, r3, c1, c2, c3, r);
        return r;

}                                

static void _C_PyMatrix_InvertTo(double *m, double *i) {

	double det;
	int j;

        i[0] = det3(m, 1, 2, 3, 1, 2, 3);
        i[1] = -det3(m, 1, 2, 3, 0, 2, 3);
        i[2] = det3(m, 1, 2, 3, 0, 1, 3);
        i[3] = -det3(m, 1, 2, 3, 0, 1, 2);
        
        i[4] = -det3(m, 0, 2, 3, 1, 2, 3);
        i[5] = det3(m, 0, 2, 3, 0, 2, 3);
        i[6] = -det3(m, 0, 2, 3, 0, 1, 3);
        i[7] = det3(m, 0, 2, 3, 0, 1, 2);
        
        i[8] = det3(m, 0, 1, 3, 1, 2, 3);
        i[9] = -det3(m, 0, 1, 3, 0, 2, 3);
        i[10] = det3(m, 0, 1, 3, 0, 1, 3);
        i[11] = -det3(m, 0, 1, 3, 0, 1, 2);

        i[12] = -det3(m, 0, 1, 2, 1, 2, 3);
        i[13] = det3(m, 0, 1, 2, 0, 2, 3);
        i[14] = -det3(m, 0, 1, 2, 0, 1, 3);
        i[15] = det3(m, 0, 1, 2, 0, 1, 2);

        det = m[0]*i[0] + m[4]*i[1] + m[8]*i[2] + m[12]*i[3];
        
        for (j = 0; j < 16; j++)
                i[j] /= det;
}        


PyObject *C_PyMatrix_Invert(PyObject *self) {

        PyMatrix *mm = (PyMatrix *)self;
	PyMatrix *im = (PyMatrix *)C_PyMatrix_New();
	double *m = mm->values;
        double *i = im->values;

        _C_PyMatrix_InvertTo(m, i);
        
        return (PyObject *)im;

}

static PyObject *PyMatrix_Invert(PyObject *self, PyObject *args)
{
        return C_PyMatrix_Invert(self);
}

static PyMethodDef PyMatrix_methods[] = {
        {"multiply", (PyCFunction)PyMatrix_Multiply, METH_VARARGS,
                "Multiply two matrices together"},
        {"invert", (PyCFunction)PyMatrix_Invert, METH_VARARGS,
                "Return the inverse of a matrix"},
	{"multiplyVector", (PyCFunction)PyMatrix_MultiplyVector, METH_VARARGS,
		"Multiple the matrix by a vector"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyMatrix_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyMatrix_methods, self, name);
        
}

static int PyMatrix_Print(PyObject *self, FILE *fp, int flags) {

        C_PyMatrix_Print(self, fp, flags);
        return 0;

}

static PyTypeObject PyMatrixType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyMatrix",
        sizeof(PyMatrix),
        0,
        C_PyMatrix_Free,
        PyMatrix_Print, /* print */
        PyMatrix_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
