/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

/*
typedef struct {
        void (*generateTransform)(PyObject *self, PyObject *status, int time, 
                        int gen_forward, int gen_reverse, int apply);
        void (*transformByMe)(PyObject *self);
        void (*invertByMe)(PyObject *self);
        int (*registerSize)(PyObject *self);
        void (*registerFunc)(PyObject *self, PyObject *status, int delay,
                        PyObject *callback, PyObject *user, PyObject *user2);
        void (*deRegisterFunc)(PyObject *self, PyObject *status);
} ITransform;
*/
#include "TograIncludes.h"
staticforward PyTypeObject PyODETransformType;

static void C_PyODETransform_GenerateTransform(PyObject *self, PyObject *status,
                int time, int gen_forward, int gen_reverse, int apply);
inline void C_PyODETransform_TransformByMe(PyObject *self);
inline void C_PyODETransform_InvertByMe(PyObject *self);
static int C_PyODETransform_RegisterSize(PyObject *self) {return 0;}

static ITransform PyODETransform_ITransform = {C_PyODETransform_GenerateTransform,
        C_PyODETransform_TransformByMe, C_PyODETransform_InvertByMe, 
        C_PyODETransform_RegisterSize, NULL, NULL};

static IC PyODETransform_Interfaces = {NULL, NULL, NULL, NULL, NULL, 
        &PyODETransform_ITransform};

static inline int PyODETransform_Check(PyObject *v) {

        return v->ob_type == &PyODETransformType;

}

PyObject *C_PyODETransform_New(PyObject *om) {
		
	PyODETransform *t = PyObject_New(PyODETransform, &PyODETransformType);

        D_ALLOC(t, "Transform");
        
	t->interfaces = &PyODETransform_Interfaces;
	t->mixin = (PyODEMixin *)om;
		
	t->body = dBodyCreate(t->mixin->world);

	Py_INCREF(om);
	
	return (PyObject *)t;
	
}

PyObject *PyODETransform_New(PyObject *self, PyObject *args) {

	PyObject *mixin;
	
	if (!PyArg_ParseTuple(args, "O", &mixin)) {
		return NULL;
	}
	
	return C_PyODETransform_New(mixin);
	
}

static void C_PyODETransform_Free(PyObject *self) {
	
        D_FREE((PyODETransform *)self, "Transform");

	PyODETransform *t = (PyODETransform *)self;

	Py_DECREF(t->mixin);

	PyObject_Del(self);

	/*
	 * TODO: Free body
	 */
	
}

static void PyODETransform_Free(PyObject *self) {
	
	C_PyODETransform_Free(self);
	
}

static void C_PyODETransform_SetSphere(PyObject *self, double density, 
		double radius) {

	PyODETransform *t = (PyODETransform *)self;

	dMass m;
	t->geom = dCreateSphere(t->mixin->space, radius);
	
	dMassSetSphere(&m, density, radius);
	dBodySetMass(t->body, &m);
	dGeomSetBody(t->geom, t->body);

}

static void C_PyODETransform_SetBox(PyObject *self, double density,
		double x_dim, double y_dim, double z_dim) {

	PyODETransform *t = (PyODETransform *)self;

	dMass m;
	t->geom = dCreateBox(t->mixin->space, x_dim, y_dim, z_dim);
	
	dMassSetBox(&m, density, x_dim, y_dim, z_dim);
	dBodySetMass(t->body, &m);
	dGeomSetBody(t->geom, t->body);
}

static PyObject *PyODETransform_SetSphere(PyObject *self, PyObject *args) {

	double d, r;

	if (!PyArg_ParseTuple(args, "dd", &d, &r)) {
		return NULL;
	}

	C_PyODETransform_SetSphere(self, d, r);

	Py_INCREF(Py_None);
	return Py_None;
	
}

static PyObject *PyODETransform_SetBox(PyObject *self, PyObject *args) {

	double d, x, y, z;

	if (!PyArg_ParseTuple(args, "dddd", &d, &x, &y, &z)) {
		return NULL;
	}

	C_PyODETransform_SetBox(self, d, x, y, z);

	Py_INCREF(Py_None);
	return Py_None;
	
}
static void C_PyODETransform_SetPosition(PyObject *self, PyObject *translation) {
	
	PyODETransform *t = (PyODETransform *)self;
	
	double x, y, z;

	x = ((PyVector *)translation)->vector[0];
	y = ((PyVector *)translation)->vector[1];
	z = ((PyVector *)translation)->vector[2];

	dBodySetPosition(t->body, x, y, z);

}

static PyObject *PyODETransform_SetPosition(PyObject *self, PyObject *args) {
	
	PyObject *arg;
	
	if (!PyArg_ParseTuple(args, "O", &arg)) {
		return NULL;
	}
	
	C_PyODETransform_SetPosition(self, arg);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}

PyObject *C_PyODETransform_GetPosition(PyObject *self) {
	
	const double *p;
	PyODETransform *t = (PyODETransform *)self;

	p = dBodyGetPosition(t->body);

	return C_PyVector_FromValues(p[0], p[1], p[2]);
}

static PyObject *PyODETransform_GetPosition(PyObject *self, PyObject *args) {

	return C_PyODETransform_GetPosition(self);

}

static PyObject *C_PyODETransform_GetForce(PyObject *self) {
	
	const double *p;
	PyODETransform *t = (PyODETransform *)self;

	p = dBodyGetForce(t->body);

	return C_PyVector_FromValues(p[0], p[1], p[2]);
}

static PyObject *PyODETransform_GetForce(PyObject *self, PyObject *args) {

	return C_PyODETransform_GetForce(self);

}

PyObject *C_PyODETransform_GetVelocity(PyObject *self) {
	
	const double *p;
	PyODETransform *t = (PyODETransform *)self;

	p = dBodyGetLinearVel(t->body);

	return C_PyVector_FromValues(p[0], p[1], p[2]);
}

static PyObject *PyODETransform_GetVelocity(PyObject *self, PyObject *args) {

	return C_PyODETransform_GetVelocity(self);

}


static void C_PyODETransform_SetVelocity(PyObject *self, PyObject *translation) {
	
	PyODETransform *t = (PyODETransform *)self;
	
	double x, y, z;

	x = ((PyVector *)translation)->vector[0];
	y = ((PyVector *)translation)->vector[1];
	z = ((PyVector *)translation)->vector[2];

	dBodySetLinearVel(t->body, x, y, z);

}

static PyObject *PyODETransform_SetVelocity(PyObject *self, PyObject *args) {
	
	PyObject *arg;
	
	if (!PyArg_ParseTuple(args, "O", &arg)) {
		return NULL;
	}
	
	C_PyODETransform_SetVelocity(self, arg);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}

void C_PyODETransform_AddForce(PyObject *self, double x, double y, 
		double z)
{

	PyODETransform *t = (PyODETransform *)self;

	dBodyAddForce(t->body, x, y, z);

}

static PyObject *PyODETransform_AddForce(PyObject *self, PyObject *args) {

	PyObject *arg;
	double x, y, z;

	if (!PyArg_ParseTuple(args, "O", &arg)) {
		return NULL;
	}


	x = ((PyVector *)arg)->vector[0];
	y = ((PyVector *)arg)->vector[1];
	z = ((PyVector *)arg)->vector[2];

	C_PyODETransform_AddForce(self, x, y, z);

	Py_INCREF(Py_None);
	return Py_None;
}

static void C_PyODETransform_SetForce(PyObject *self, PyObject *force)
{

	PyODETransform *t = (PyODETransform *)self;

	double x, y, z;

	x = ((PyVector *)force)->vector[0];
	y = ((PyVector *)force)->vector[1];
	z = ((PyVector *)force)->vector[2];

	dBodySetForce(t->body, x, y, z);

}

static PyObject *PyODETransform_SetForce(PyObject *self, PyObject *args) {

	PyObject *arg;

	if (!PyArg_ParseTuple(args, "O", &arg)) {
		return NULL;
	}

	C_PyODETransform_SetForce(self, arg);

	Py_INCREF(Py_None);
	return Py_None;
}

/*
static void C_PyODETransform_SetRotationAxis(PyObject *self, PyObject *axis) {
	
	PyODETransform *t = (PyODETransform *)self;
	
	Py_DECREF(t->rotationAxis);
	if (axis != NULL)
		Py_INCREF(axis);
	t->rotationAxis = axis;
	
}

static PyObject *C_PyODETransform_GetRotationAxis(PyObject *self) {

	PyODETransform *t = (PyODETransform *)self;
	
	if (t->rotationAxis != NULL)
	{
		Py_INCREF(t->rotationAxis);
	}
	
	return t->rotationAxis;
}

static PyObject *PyODETransform_GetRotationAxis(PyObject *self, PyObject *args) {

	PyObject *ret = C_PyODETransform_GetRotationAxis(self);

	if (ret == NULL) {
		Py_INCREF(Py_None);
		return Py_None;
	} else {
		return ret;
	}
}
	
static PyObject *PyODETransform_SetRotationAxis(PyObject *self, PyObject *args) {
	
	PyObject *arg;
	
	if (!PyArg_ParseTuple(args, "O", &arg)) {
		return NULL;
	}

	if (arg == Py_None)
		arg = NULL;
	
	C_PyODETransform_SetRotationAxis(self, arg);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}

static void C_PyODETransform_SetRotationAmount(PyObject *self, double amount) {
	
	PyODETransform *t = (PyODETransform *)self;
	
	t->rotationAmount = amount;
	
}

static double C_PyODETransform_GetRotationAmount(PyObject *self) {

	PyODETransform *t = (PyODETransform *)self;

	return t->rotationAmount;

}

static PyObject *PyODETransform_SetRotationAmount(PyObject *self, PyObject *args) {
	
	double val;
	if (!PyArg_ParseTuple(args, "d", &val)) {
		return NULL;
	}
	
	C_PyODETransform_SetRotationAmount(self, val);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}

static PyObject *PyODETransform_GetRotationAmount(PyObject *self, PyObject *args) {

	double val = C_PyODETransform_GetRotationAmount(self);

	return PyFloat_FromDouble(val);

}

static void C_PyODETransform_SetScale(PyObject *self, PyObject *scale) {
	
	PyODETransform *t = (PyODETransform *)self;
	
	Py_DECREF(t->scale);
	Py_INCREF(scale);
	t->scale = scale;
	
}

static PyObject *PyODETransform_SetScale(PyObject *self, PyObject *args) {
	
	PyObject *arg;
	
	if (!PyArg_ParseTuple(args, "O", &arg)) {
		return NULL;
	}
	
	C_PyODETransform_SetScale(self, arg);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}
*/

inline void C_PyODETransform_TransformByMe(PyObject *self) {

	PyODETransform *t = (PyODETransform *)self;
	const double *pos = dGeomGetPosition(t->geom);
	const double *rot = dGeomGetRotation(t->geom);

	double matrix[16];
	int i, j;
	
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			matrix[j + i * 4] = rot[j * 4 + i];
		}
		matrix[i * 4 + 3] = 0;
	}

	for (i = 0; i < 3; i++) {
		matrix[12 + i] = pos[i];
	}

	matrix[15] = 1;

	/*
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < 4; j++)
		{
			printf("%2.2f, ", matrix[i * 4 + j]);
		}
		printf("\n");
	}
	printf("\n");
	*/

	glMultMatrixd(matrix);
	//glTranslated(pos[0], pos[1], pos[2]);

	/*
	if (v->vector[0] > 1)
	{
#if DEBUG_VERBOSE
		printf("scale: (%.2f,%.2f,%.2f)\n", v->vector[0], v->vector[1],
        	                v->vector[2]);
#endif

	        glScaled(v->vector[0], v->vector[1], v->vector[2]);
	}
        if (t->rotationAmount != 0)
        {
                v = (PyVector *)t->rotationAxis;
#if DEBUG_VERBOSE		
                printf("rotate: %.2f @ (%.2f,%.2f,%.2f)\n", t->rotationAmount, 
                                v->vector[0], v->vector[1], v->vector[2]);
#endif
	        glRotated(t->rotationAmount, v->vector[0], v->vector[1], 
                                v->vector[2]);
        }
	*/
	
}

inline void C_PyODETransform_InvertByMe(PyObject *self) {

        PyODETransform *t = (PyODETransform *)self;
	
	const double *pos = dGeomGetPosition(t->geom);
	glTranslated(-pos[0], -pos[1], -pos[2]);

}

static void C_PyODETransform_GenerateTransform(PyObject *self, PyObject *status,
                int time, int gen_forward, int gen_reverse, int apply)
{

        if (apply == 1)
                C_PyODETransform_TransformByMe(self);
        else if (apply == -1)
                C_PyODETransform_InvertByMe(self);

}
        

static PyMethodDef PyODETransform_methods[] = {
        {"setPosition", (PyCFunction)PyODETransform_SetPosition,METH_VARARGS,
                "Sets the position of this transform's ODE object"},
        {"getPosition", (PyCFunction)PyODETransform_GetPosition,METH_VARARGS,
                "Gets the position of this transform's ODE object"},
        {"setVelocity", (PyCFunction)PyODETransform_SetVelocity,METH_VARARGS,
                "Sets the velocity of this transform's ODE object"},
        {"getVelocity", (PyCFunction)PyODETransform_GetVelocity,METH_VARARGS,
                "Gets the velocity of this transform's ODE object"},
        {"addForce", (PyCFunction)PyODETransform_AddForce,METH_VARARGS,
                "Add a force to this transform's ODE object"},
        {"setForce", (PyCFunction)PyODETransform_SetForce,METH_VARARGS,
                "Set the force on this transform's ODE object"},
        {"getForce", (PyCFunction)PyODETransform_GetForce,METH_VARARGS,
                "Get the force on this transform's ODE object"},
        {"setSphere", (PyCFunction)PyODETransform_SetSphere, 
                METH_VARARGS, "Set this ODE object to be a sphere"},
        {"setBox", (PyCFunction)PyODETransform_SetBox, 
                METH_VARARGS, "Set this ODE object to be a box"},
/*
	{"getRotationAxis", (PyCFunction)PyODETransform_GetRotationAxis, 
                METH_VARARGS, "Get the rotation axis of this transform"},
        {"setRotationAmount", (PyCFunction)PyODETransform_SetRotationAmount,
                METH_VARARGS, "Set the angle of rotation for this transform"},
        {"getRotationAmount", (PyCFunction)PyODETransform_GetRotationAmount,
                METH_VARARGS, "Get the angle of rotation for this transform"},
        {"setScale", (PyCFunction)PyODETransform_SetScale, METH_VARARGS,
                "Sets the scale vector of this transform"},
*/
        {NULL, NULL, 0, NULL}
};

static PyObject *PyODETransform_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyODETransform_methods, self, name);

}

static PyTypeObject PyODETransformType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyODETransform",
        sizeof(PyODETransform),
        0,
        PyODETransform_Free,
        0, /* print */
        PyODETransform_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
