/*
 * Copyright 2003, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

TOGRA_DEFINE_MIXIN(ProtoMixin)

void *C_PyProtoMixin_New(void) {

	PyProtoMixin *om = PyObject_New(PyProtoMixin, &PyProtoMixinType);
	D_ALLOC(om, "ProtoMixin");

	om->interfaces = &PyProtoMixin_Interfaces;
	om->Init = Py_None;
	om->Destroy = Py_None;
	om->RenderStep = Py_None;

	Py_INCREF(Py_None);
	Py_INCREF(Py_None);
	Py_INCREF(Py_None);

	return (void *)om;

}

PyObject *PyProtoMixin_New(PyObject *self, PyObject *args) {

	return C_PyProtoMixin_New();

}

DEF_PROTO_FUNCTION1(ProtoMixin, Init, int, time, "i")
DEF_PROTO_FUNCTION1(ProtoMixin, Destroy, int, time, "i")
DEF_PROTO_FUNCTION3(ProtoMixin, RenderStep,int,time,int,is_paused,int,paused_time, "iii")

static void PyProtoMixin_Free(PyObject *self) {

	D_FREE(self, "ProtoMixin");

}

SET_PROTO_FUNCTION(ProtoMixin, Init)
SET_PROTO_FUNCTION(ProtoMixin, Destroy)
SET_PROTO_FUNCTION(ProtoMixin, RenderStep)

static PyMethodDef PyProtoMixin_methods[] = {
	{"setInit", (PyCFunction)PyProtoMixin_SetInit, METH_VARARGS,
		"set the init function of this mixin object"},
	{"setRenderStep", (PyCFunction)PyProtoMixin_SetRenderStep, METH_VARARGS,
		"set the renderStep function of this mixin object"},
	{"setDestroy", (PyCFunction)PyProtoMixin_SetDestroy, METH_VARARGS,
		"set the destroy function of this mixin object"},
	{NULL, NULL, 0, NULL}
};

static PyObject *PyProtoMixin_GetAttr(PyObject *self, char *name) {
	return Py_FindMethod(PyProtoMixin_methods, self, name);
}

TOGRA_DEFINE_GETATTR_OBJECT(ProtoMixin)
