/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

#include "Vector_Privates.h"

/* empty comment */

staticforward PyTypeObject PyStaticSurfaceType;

static void C_PyStaticSurface_Render(PyObject *self, int type, int picking);
static int C_PyStaticSurface_Compile(PyObject *self);

static IRender PyStaticSurface_IRender = {C_PyStaticSurface_Render, 
        C_PyStaticSurface_Compile};

static IC PyStaticSurface_Interfaces = {&PyStaticSurface_IRender, NULL};

static inline int PyStaticSurface_Check(PyObject *v) {

        return v->ob_type == &PyStaticSurfaceType;

}

static inline PyObject * PyStaticSurface_Vectors(PyObject *v) {

        return ((PyStaticSurface *)v)->vectors;

}

static inline PyObject * PyStaticSurface_Normals(PyObject *v) {

        return ((PyStaticSurface *)v)->normals;

}

void C_PyStaticSurface_Quads_GenNormals(PyObject *vectors, 
                PyObject **normals, PyObject **faces) {

        /*
        PyVectorList *v = (PyVectorList *)vectors;
        PyVectorList *n = (PyVectorList *)C_PyVectorList_OfLength(v->length);
        */
        /* unfinished ! */
}

void C_PyStaticSurface_QuadStrip_GenNormals(PyObject *vectors, 
		PyObject **normals) {
        
        PyVectorList *v = (PyVectorList *)vectors;
        PyVectorList *n = (PyVectorList *)C_PyVectorList_OfLength(v->length);
        int i;

	assert(!"GENNORMALS NOT IMPLEMENTED FOR STATICSURFACES!");
	
        for (i = 0; i + 2 < v->length; i += 2)
        {
                PyVector *v0 
                       = (PyVector *)C_PyVector_Subtract(v->pyVectors[i + 1],
                                v->pyVectors[i]);
                PyVector *v1 
                       = (PyVector *)C_PyVector_Subtract(v->pyVectors[i + 3],
                                v->pyVectors[i + 1]);

                VL_APPEND((PyObject *)n, C_PyVector_Cross((PyObject *)v0, 
                                                (PyObject *)v1));
        }

        *normals = (PyObject *)n;

}

void *C_PyStaticSurface_New(PyObject *vectors, PyObject *normals, 
                PyObject *textureCoords, int textureMode, int width, 
		int height, int normalMode, int genNormals) {

        PyStaticSurface *r 
		= PyObject_New(PyStaticSurface, &PyStaticSurfaceType);

        D_ALLOC(r, "StaticSurface");

        r->interfaces = &PyStaticSurface_Interfaces;
        r->vectors = vectors;
        r->normals = normals;
        r->width = width;
        r->height = height;
        r->normal_mode = normalMode;
        r->display_list = 0;
        r->textureCoords = textureCoords;
        r->textureMode = textureMode;

	if (vectors != NULL)
	        Py_INCREF(vectors);

        if (normals != NULL)
                Py_INCREF(normals);

        if (textureCoords != NULL)
                Py_INCREF(textureCoords);
        
        if (normals != NULL || (!genNormals) || normalMode == 0)
                return (void *)r;

        /*
         * Not all possibilities implemented
         */
	C_PyStaticSurface_QuadStrip_GenNormals(vectors, &(r->normals));

	
        return (void *)r;
        
}

static void C_PyStaticSurface_FlatSurface(PyObject *self, double x0, 
		double y0, double x1, double y1) {

	PyStaticSurface *ss = (PyStaticSurface *)self;
	
	int i, j;
	double wstep = (x1 - x0)/(ss->width - 1);
	double hstep = (y1 - y0)/(ss->height - 1);

	if (ss->vectors != NULL) {
		Py_DECREF(ss->vectors);
	}
		
	if (ss->normals != NULL) {
		Py_DECREF(ss->normals);
	}
	
	ss->vectors = C_PyVectorList_OfLength(ss->width * ss->height);
	ss->normals = C_PyVectorList_OfLength(ss->width * ss->height);
	
	for (i = 0; i < ss->width; i++)
	{
		for (j = 0; j < ss->height; j++)
		{
			VL_APPEND(ss->vectors, 
				C_PyVector_FromValues(
					x0 + i * wstep, y0 + j * hstep, 0));
			VL_APPEND(ss->normals,
				C_PyVector_FromValues(0,0,1));
		}
	}
}

static void C_PyStaticSurface_WobbleSurface(PyObject *self, double x0, 
		double y0, double x1, double y1, double xper, double yper,
		double xmag, double ymag) {

	int i, j;
	PyStaticSurface *ss = (PyStaticSurface *)self;
	double *pos;
	
	C_PyStaticSurface_FlatSurface(self, x0, y0, x1, y1);
	pos = ((PyVectorList *)ss->normals)->values;
	
	for (i = 0; i < ss->width; i++)
	{
		double xdiff;
		xdiff = xmag * sin(2*i*3.142/xper/(ss->width-1));
		for (j = 0; j < ss->height; j++)
		{
			double ydiff;
			ydiff = ymag * sin(2*j*3.142/yper/(ss->height-1));
			pos[0] += xdiff;
			pos[1] += ydiff;
			pos += 3;
		}
	}
}

#define FROM0_BEZIER_VAL(b,c,d,t)				\
	(3*b*(1-t)*(1-t)*t + 3*c*(1-t)*t*t + d*t*t*t)
/*
 * 3b(1-t)(1-t)t + 3c(1-t)tt + dttt ->
 * -6b(1-t)t + 3b(1-t)(1-t) - 3ctt + 6c(1-t)t + 3dtt ->
 *  3b(1-t)(1-t) + (6c-6b)(1-t)t + (3d-3c)tt
 */
#define FROM0_DBEZIER_VAL(b,c,d,t)				\
	(3*b*(1-t)*(1-t) + (6*c-6*b)*(1-t)*t + (3*d-3*c)*t*t)

#define CROSSX(x0,y0,z0,x1,y1,z1)				\
	(y0*z1-y1*z0)
#define CROSSY(x0,y0,z0,x1,y1,z1)				\
	(z0*x1-z1*x0)
#define CROSSZ(x0,y0,z0,x1,y1,z1)				\
	(x0*y1-x1*y0)

static void C_PyStaticSurface_SplineSurface(PyObject *self, 
		double *cxa, double *cxb, double *tx,
		double *cya, double *cyb, double *ty)
{
	int i, j;

	PyStaticSurface *ss = (PyStaticSurface *)self;
	double xx[ss->width];
	double xy[ss->width];
	double xz[ss->width];
	double yx[ss->height];
	double yy[ss->height];
	double yz[ss->height];
	double dxx[ss->width];
	double dxy[ss->width];
	double dxz[ss->width];
	double dyx[ss->height];
	double dyy[ss->height];
	double dyz[ss->height];

	
	if (ss->vectors != NULL) {
		Py_DECREF(ss->vectors);
	}

	if (ss->normals != NULL) {
		Py_DECREF(ss->normals);
	}
	
	ss->vectors = C_PyVectorList_OfLength(ss->width * ss->height);
	ss->normals = C_PyVectorList_OfLength(ss->width * ss->height);
	
	/*
	printf("%f %f %f | %f %f %f | %f %f %f\n",
			cxa[0],cxb[0],tx[0],
			cxa[1],cxb[1],tx[1],
			cxa[2],cxb[2],tx[2]);
	*/
	for (i = 0; i < ss->width; i++)
	{
		double t = i*1.0/(ss->width - 1);
		xx[i] = FROM0_BEZIER_VAL(cxa[0],cxb[0],tx[0],t);
		xy[i] = FROM0_BEZIER_VAL(cxa[1],cxb[1],tx[1],t);
		xz[i] = FROM0_BEZIER_VAL(cxa[2],cxb[2],tx[2],t);
		dxx[i] = FROM0_DBEZIER_VAL(cxa[0],cxb[0],tx[0],t);
		dxy[i] = FROM0_DBEZIER_VAL(cxa[1],cxb[1],tx[1],t);
		dxz[i] = FROM0_DBEZIER_VAL(cxa[2],cxb[2],tx[2],t);
	/*
		printf("%f %f %f %f %f %f %f\n",
				xx[i],xy[i],xz[i],dxx[i],dxy[i],dxz[i], t);
	*/
	}

	for (i = 0; i < ss->height; i++)
	{
		double t = i*1.0/(ss->height - 1);
		yx[i] = FROM0_BEZIER_VAL(cya[0],cyb[0],ty[0],t);
		yy[i] = FROM0_BEZIER_VAL(cya[1],cyb[1],ty[1],t);
		yz[i] = FROM0_BEZIER_VAL(cya[2],cyb[2],ty[2],t);
		dyx[i] = FROM0_DBEZIER_VAL(cya[0],cyb[0],ty[0],t);
		dyy[i] = FROM0_DBEZIER_VAL(cya[1],cyb[1],ty[1],t);
		dyz[i] = FROM0_DBEZIER_VAL(cya[2],cyb[2],ty[2],t);
	}
	
	for (i = 0; i < ss->width; i++)
	{
		for (j = 0; j < ss->width; j++)
		{
			VL_APPEND(ss->vectors, C_PyVector_FromValues(
				xx[i]+yx[j],
				xy[i]+yy[j],
				xz[i]+yz[j]
			));
/*
			printf("(%f,%f,%f)\n", xx[i]+yx[j],xy[i]+yy[j],
					xz[i]+yz[j]);
*/
				VL_APPEND(ss->normals, C_PyVector_FromValues(
				CROSSX(dxx[i],dxy[i],dxz[i],
				       dyx[j],dyy[j],dyx[j]),
				CROSSY(dxx[i],dxy[i],dxz[i],
				       dyx[j],dyy[j],dyx[j]),
				CROSSZ(dxx[i],dxy[i],dxz[i],
				       dyx[j],dyy[j],dyx[j])
			));
		}
	}


}

PyObject *PyStaticSurface_FlatSurface(PyObject *self, PyObject *args)
{
	double x0, y0, x1, y1;
	
	if (!PyArg_ParseTuple(args, "dddd", &x0, &y0, &x1, &y1))
	{
		return NULL;
	}

	C_PyStaticSurface_FlatSurface(self, x0, y0, x1, y1);
	
	Py_INCREF(Py_None);
	return Py_None;
}

PyObject *PyStaticSurface_WobbleSurface(PyObject *self, PyObject *args)
{
	double x0, y0, x1, y1, xper, yper, xmag, ymag;
	
	if (!PyArg_ParseTuple(args, "dddddddd", &x0, &y0, &x1, &y1,
				&xper, &yper, &xmag, &ymag))
	{
		return NULL;
	}

	C_PyStaticSurface_WobbleSurface(self, x0, y0, x1, y1, xper, yper,
			xmag, ymag);
	
	Py_INCREF(Py_None);
	return Py_None;
}

PyObject *PyStaticSurface_SplineSurface(PyObject *self, PyObject *args)
{

	PyObject *cxa, *cxb, *tx, *cya, *cyb, *ty;
	
	if (!PyArg_ParseTuple(args, "OOOOOO", &cxa, &cxb, &tx, &cya, &cyb, &ty))
		return NULL;

	C_PyStaticSurface_SplineSurface(self, ((PyVector *)cxa)->vector,
		((PyVector *)cxb)->vector,((PyVector *)tx)->vector,
		((PyVector *)cya)->vector,((PyVector *)cyb)->vector,
		((PyVector *)ty)->vector);

	Py_INCREF(Py_None);
	return Py_None;

}



PyObject *PyStaticSurface_New(PyObject *self, PyObject *args, PyObject *kwords) 
{

        PyObject *vectors = NULL, *normals = NULL;
        PyObject *textureCoords = NULL;
        int width, height;
        int textureMode = 0;
        int gen_normals = 0;
        int normalMode = 0;
					
        static char *kwlist[] = {"width", "height", "vectors", "normals", 
		"normalMode", "genNormals", "textureCoords", "textureMode", 
		NULL};
        
        if (!PyArg_ParseTupleAndKeywords(args, kwords, "ii|OOiiOi",
		kwlist, &width, &height, &vectors, &normals, &normalMode,
		&gen_normals, &textureCoords, &textureMode)) { 
		return NULL; 
	}

        /* TODO: Implement check for length of normals */
        
        return (PyObject *)C_PyStaticSurface_New(vectors, normals, 
                        textureCoords, textureMode, width, height, normalMode, 
                        gen_normals);
}

static void PyStaticSurface_Free(PyObject *self) {

        D_FREE(self, "StaticSurface");
        
        /* TODO */
        
}

void C_PyStaticSurface_SetTextureCoords(PyObject *obj, PyObject *v) {

        PyStaticSurface *r = (PyStaticSurface *)obj;
        
        if (r->textureCoords != NULL)
        {
                Py_DECREF(r->textureCoords);
        }
        
        Py_INCREF(v);
        r->textureCoords = v;


}

static int PyStaticSurface_Setattr(PyObject *obj, char *name, PyObject *v) {
	
	/*PyStaticSurface *r = (PyStaticSurface *)obj;*/

	if (strcmp(name, "textureCoords") == 0) {

                C_PyStaticSurface_SetTextureCoords(obj, v);
                return 0;
                  
        }
	
	return -1;
}

static PyMethodDef PyStaticSurface_methods[] = {
	{"flatSurface", (PyCFunction)PyStaticSurface_FlatSurface,
		METH_VARARGS, "Generate a flat surface"},
	{"wobbleSurface", (PyCFunction)PyStaticSurface_WobbleSurface,
		METH_VARARGS, "Generate a wobble surface"},
	{"splineSurface", (PyCFunction)PyStaticSurface_SplineSurface,
		METH_VARARGS, "Generate a spline surface"},
	{NULL, NULL, 0, NULL}
};

static PyObject *PyStaticSurface_Getattr(PyObject *obj, char *name) {
	
	PyStaticSurface *r = (PyStaticSurface *)obj;

        if (strcmp(name, "textureCoords") == 0) {

                if (r->textureCoords == NULL) {
                        Py_INCREF(Py_None);
                        return Py_None;
                }

                Py_INCREF(r->textureCoords);
                return r->textureCoords;

        }
	
	return Py_FindMethod(PyStaticSurface_methods, obj, name);
	
}

inline void _texture(int mode, float *texture, int pos);
/*
{

        if (texture == NULL)
                return;

        if (mode == 0)
        {
                glTexCoord2f(texture[pos*2], texture[pos*2+1]);
        }
}
*/              

static void C_StaticSurface_Lists_Static(double *v, double *n,  
                float *texture, int width, int height, int normalMode, 
                int tMode) {

        int i;

        if (normalMode == 1)
        {

                int n_pos = 0;
                assert(n != NULL);
		
		/*
		 * 0   1   2   3   4
		 *   a   b   c   d
		 * 5   6   7   8   9
		 *   e   f   g   h
		 * 10  11  12  13  14
		 *
		 * So:   0,5,6,1     1,6,7,2     2,7,8,3     3,8,4,9
		 * and:  5,10,11,6   6,11,12,7   7,12,13,8   8,13,14,9
		 *
		 * width = 3, height = 5
		 * 
		 * for i in range(width - 1):
		 * 	for j in range(height - 1):
		 * 		V(j+height*i)
		 * 		V(j+height*i+1)
		 * 		V(j+height*(i+1)+1)
		 * 		V(j+height*(i+1))
		 */
		 
		for (i = 0; i < width - 1; i++)
		{
			int j;
                
			glBegin(GL_QUADS);

			for (j = 0; j < height - 1; j++)
			{
				int pos = j + height * i;
				glNormal3dv(&n[(n_pos++)*3]);
				_texture(tMode, texture, pos);
				glVertex3dv(&v[pos*3]);
				pos++;
				_texture(tMode, texture, pos);
				glVertex3dv(&v[pos*3]);
				pos += height;
				_texture(tMode, texture, pos);
				glVertex3dv(&v[pos*3]);
				pos--;
				_texture(tMode, texture, pos);
				glVertex3dv(&v[pos*3]);
			}

			glEnd();
		}	

                return;
                
        }
      
	/*
	 * Otherwise: 
	 * 0   1   2   3   4
	 *
	 * 5   6   7   8   9
	 *
	 * 10  11  12  13  14
	 *
	 * So:   0,5,1,6,2,7,3,8,4,9
	 * and:  5,10,6,11,7,12,8,13,9,14
	 *
	 * width = 3, height = 5
	 * 
	 * for i in range(width - 1):
	 * 	for j in range(height):
	 * 		V(j+height*i)
	 * 		V(j+height*(i+1))
	 */

	for (i = 0; i < width - 1; i++)
	{
		int j;
	        glBegin(GL_QUAD_STRIP);

        
		for (j = 0; j < height; j++) {
	
			if (n != NULL) {
				glNormal3dv(&n[(j+height*i)*3]);
			}
	
			_texture(tMode, texture, (j+height*i));
			glVertex3dv(&v[(j+height*i)*3]);
	
			/*
			PRINT_FV("place 1", vt);
			*/
			
			if (n != NULL) {
				glNormal3dv(&n[(j+height*(i+1))*3]);
			}
	
			_texture(tMode, texture, (j+height*(i+1)));
			glVertex3dv(&v[(j+height*(i+1))*3]);
			
			/*
			PRINT_FV("place 2", vt2);
			*/
		}
        	glEnd();
	}

                        
}

static int C_PyStaticSurface_Compile(PyObject *self) {

        double *v, *n = NULL;
        float *t = NULL;
        
        PyStaticSurface *r = (PyStaticSurface *)self;

#if DEBUG_VERBOSE
	printf("Compiling %p\n", self);
#endif
        if (r->display_list == 0)
        {
        
                v = VL_VALUES(r->vectors);
                if (r->normals != NULL) {
                        n = VL_VALUES(r->normals);
                }

		r->display_list = glGenLists(1);

                if (r->textureCoords != NULL) {
                        t = ((PyFloatList *)r->textureCoords)->floats;
                }
                
                if (r->display_list == 0)
                {
                        /* FIXME: Is this the correct behaviour? */
                        return FALSE;
                }
                glNewList(r->display_list, GL_COMPILE);

                C_StaticSurface_Lists_Static(v, n, t, r->width, r->height,
                                r->normal_mode, r->textureMode);
                glEndList();
        }

        return TRUE;
                
}

static void C_PyStaticSurface_Render(PyObject *self, int time, int picking) {

        PyStaticSurface *r = (PyStaticSurface *)self;

        if (r->display_list == 0) {
		TPRINTF("Compiling %p\n", self);
                C_PyStaticSurface_Compile(self);
        }
	TPRINTF("Calling %p\n", r);

        glCallList(r->display_list);
}

static PyObject *PyStaticSurface_Call(PyObject *self, PyObject *args, PyObject *dict) {

        PyObject *p;
        int gl_type = GL_POINTS;


        if (!(PyTuple_Size(args) == 1)) {
                return NULL;
        }

        p = PyTuple_GetItem(args, 0);

        if (p != Py_None) {

                if (!(PyInt_Check(p))) {
                        return NULL;
                }

                gl_type = (int) PyInt_AsLong(p);

        }

        C_PyStaticSurface_Render((void *)self, 0, 0);

        Py_INCREF(Py_None);
        return Py_None;

}
        
static PySequenceMethods PyStaticSurface_sequenceMethods = {
        0, /* length */
        0, /* concat */
        0, /* repeat */
        0, /* item */
        0, /* slice */
        0, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyStaticSurfaceType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyStaticSurface",
        sizeof(PyStaticSurface),
        0,
        PyStaticSurface_Free,
        0, /* print */
        PyStaticSurface_Getattr, /* getattr */
        PyStaticSurface_Setattr, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyStaticSurface_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        &PyStaticSurface_Call, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
