/*
 * Python object representing a GL StaticTexture
 */

#include "TograIncludes.h"
staticforward PyTypeObject PyStaticTextureType;

static void C_PyStaticTexture_SetStaticTexture(PyObject *material, int time);

static ITexture PyStaticTexture_IStaticTexture = 
	{C_PyStaticTexture_SetStaticTexture};
static IC PyStaticTexture_Interfaces = {NULL, NULL, NULL, NULL, NULL, NULL, 
	NULL, &PyStaticTexture_IStaticTexture};

PyObject *C_PyStaticTexture_New(int *t, int w, int h, int format) {

        PyStaticTexture *m 
		= (PyStaticTexture *)PyObject_New(PyStaticTexture, 
				&PyStaticTextureType);

	
	D_ALLOC(m, "StaticTexture");

        m->interfaces = &PyStaticTexture_Interfaces;
	
	m->format = GL_RGBA;

        if (format != -1)
                m->format = format;
        
        glGenTextures(1, &(m->texture));
        glBindTexture(GL_TEXTURE_2D, m->texture);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
 
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, m->format, 
                        GL_UNSIGNED_BYTE, (void *)t);

        return (PyObject *)m;
        
}

PyObject *PyStaticTexture_New(PyObject *self, PyObject *args) {

        PyObject *texture, *t;
        int width, height, format = -1;
        
        if (!PyArg_ParseTuple(args, "Oii|i", &texture, &width, &height,
                                &format)) {
                return NULL;
        }

        /* TODO: This needs to be converted to check for IL INTERFACE */
        
        if (PyIntList_Check(texture)) {
                t = C_PyStaticTexture_New(IL_INTS(texture), width, 
                                height, format);
        } else if (PyObject_TypeCheck(texture, &PyString_Type)) {
                t = C_PyStaticTexture_New((int *)PyString_AsString(texture),
                                width, height, format);
        } else {
		t = NULL;
	}

        return t;
        
}

static void PyStaticTexture_Free(PyObject *material) {

        D_FREE(material, "StaticTexture");
        
}

#define PFLOATPARM(s,t)	\
	printf("%s (%2.2f,%2.2f,%2.2f,%2.2f)\n", s, t[0], t[1], t[2], t[3]);

static void C_PyStaticTexture_SetStaticTexture(PyObject *m, int time) {

	
	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ONE);
	glBindTexture(GL_TEXTURE_2D, ((PyStaticTexture *)m)->texture);
        
	
}

static PyTypeObject PyStaticTextureType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyContainer",
        sizeof(PyStaticTexture),
        0,
        PyStaticTexture_Free,
        0, /* print */
        0, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
