/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"
#include "math.h"

#include "Container_Privates.h"

PyObject *C_TograUnitSphere(int slices, int segs) {
	
	double slice_height = 2.0 / slices;
	double seg_angle = 2 * M_PI / segs;
	int slice, seg;
	double cur_slice;
	double cur_r;
	double last_slice;
	double last_r;
	
	PyObject *top_slice = C_PyVectorList_OfLength(segs + 2);
	PyObject *bot_slice = C_PyVectorList_OfLength(segs + 2);
	PyObject *container;

	VL_APPEND(top_slice, C_PyVector_FromValues(0.0, 1.0, 0.0));
	VL_APPEND(bot_slice, C_PyVector_FromValues(0.0, -1.0, 0.0));	
	
	cur_slice = 1 - slice_height;
	cur_r = sqrt(1 -  cur_slice * cur_slice);

        for (seg = 0; seg <= segs; seg++) {
		VL_APPEND(top_slice, C_PyVector_FromValues(
			cur_r * sin(seg * seg_angle), 
			cur_slice, 
			cur_r * cos(seg * seg_angle)));
		VL_APPEND(bot_slice, C_PyVector_FromValues(
			cur_r * sin((segs - seg) * seg_angle), 
			-cur_slice, 
			cur_r * cos((segs - seg) * seg_angle)));
	}		

        container = C_PySphere_New();
        ((PySphere *)container)->slices = slices;
        ((PySphere *)container)->segs = segs;

        C_PyContainer_Append(container,
                C_PyStaticRender_New(top_slice, GL_TRIANGLE_FAN, top_slice, 
                                NULL, NULL, 0, segs + 2, 0, 0));

	for (slice = 1; slice < slices - 1; slice++) {
		
		PyObject *sl_obj = C_PyVectorList_OfLength(2 * segs + 2);
		
		last_slice = cur_slice;
		last_r = cur_r;
		
		cur_slice -= slice_height;
		cur_r = sqrt(1 - cur_slice * cur_slice);
		
		for (seg = segs; seg >= 0; seg--) {
                        if (cur_slice >= 0) {
                                VL_APPEND(sl_obj, C_PyVector_FromValues(
                                        cur_r * sin(seg * seg_angle),
                                        cur_slice,
                                        cur_r * cos(seg * seg_angle)));
                                VL_APPEND(sl_obj, C_PyVector_FromValues(
                                        last_r * sin(seg * seg_angle),
                                        last_slice,
                                        last_r * cos(seg * seg_angle)));
                        } else {
                                VL_APPEND(sl_obj, C_PyVector_FromValues(
                                        cur_r * sin(seg * seg_angle),
                                        cur_slice,
                                        cur_r * cos(seg * seg_angle)));
                                VL_APPEND(sl_obj, C_PyVector_FromValues(
                                        last_r * sin(seg * seg_angle),
                                        last_slice,
                                        last_r * cos(seg * seg_angle)));
                        }
		}
			
		C_PyContainer_Append(container,
                        C_PyStaticRender_New(sl_obj, GL_QUAD_STRIP, sl_obj, 
                                NULL, NULL, 0, 2 * segs + 2, 0, 0));
	}

        
	C_PyContainer_Append(container, 
                C_PyStaticRender_New(bot_slice, GL_TRIANGLE_FAN, bot_slice, 
                                NULL, NULL, 0, segs + 2, 0, 0));

	
	return container;

	
}

void C_PyTransformContainer_Append(PyObject *tc, PyObject *transform, 
		void *data);

PyObject *C_TograSphere(int cx, int cy, int cz, double r, int slices, int segs) {

	PyContainer *container = (PyContainer *)C_TograUnitSphere(slices, segs);

        C_PyTransformContainer_Append(container->transform,
                C_PyTransform_New(C_PyVector_FromValues(cx, cy, cz), 
                        C_PyVector_New(), 0, C_PyVector_FromValues(r, r, r)),
                        NULL);

	//C_PyContainer_Append(container, unitSphere);
	return (PyObject *)container;
	
}

PyObject *C_TograUnitBox(void) {
	
	PyObject *vl = C_PyVectorList_OfLength(24);
	PyObject *nl = C_PyVectorList_OfLength(24);
	
	VL_APPEND(vl, C_PyVector_FromValues(0.5,-0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(1,0,0));
	VL_APPEND(vl, C_PyVector_FromValues(0.5,-0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(1,0,0));
	VL_APPEND(vl, C_PyVector_FromValues(0.5,0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(1,0,0));
	VL_APPEND(vl, C_PyVector_FromValues(0.5,0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(1,0,0));
	
	VL_APPEND(vl, C_PyVector_FromValues(-0.5,-0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(-1,0,0));
	VL_APPEND(vl, C_PyVector_FromValues(-0.5,-0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(-1,0,0));
	VL_APPEND(vl, C_PyVector_FromValues(-0.5,0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(-1,0,0));
	VL_APPEND(vl, C_PyVector_FromValues(-0.5,0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(-1,0,0));	

	VL_APPEND(vl, C_PyVector_FromValues(-0.5,-0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,0,1));
	VL_APPEND(vl, C_PyVector_FromValues(0.5,-0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,0,1));
	VL_APPEND(vl, C_PyVector_FromValues(0.5,0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,0,1));	
	VL_APPEND(vl, C_PyVector_FromValues(-0.5,0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,0,1));

	VL_APPEND(vl, C_PyVector_FromValues(0.5,-0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,0,-1));
	VL_APPEND(vl, C_PyVector_FromValues(-0.5,-0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,0,-1));
	VL_APPEND(vl, C_PyVector_FromValues(-0.5,0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,0,-1));
	VL_APPEND(vl, C_PyVector_FromValues(0.5,0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,0,-1));	

	VL_APPEND(vl, C_PyVector_FromValues(-0.5,-0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,-1,0));
	VL_APPEND(vl, C_PyVector_FromValues(-0.5,-0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,-1,0));
	VL_APPEND(vl, C_PyVector_FromValues(0.5,-0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,-1,0));
	VL_APPEND(vl, C_PyVector_FromValues(0.5,-0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,-1,0));	

	VL_APPEND(vl, C_PyVector_FromValues(-0.5,0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,1,0));
	VL_APPEND(vl, C_PyVector_FromValues(-0.5,0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,1,0));
	VL_APPEND(vl, C_PyVector_FromValues(0.5,0.5,0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,1,0));	
	VL_APPEND(vl, C_PyVector_FromValues(0.5,0.5,-0.5));
	VL_APPEND(nl, C_PyVector_FromValues(0,1,0));

	return C_PyStaticRender_New(vl, GL_QUADS, nl, NULL, NULL, 0, 24, 0, 0);
	
}

PyObject *C_TograBox(double cx, double cy, double cz, 
					 double dx, double dy, double dz) {
	
	PyObject *unitBox = C_TograUnitBox();
	PyContainer *container = (PyContainer *)C_PyContainer_OfLength(1);

        C_PyTransformContainer_Append(container->transform, C_PyTransform_New
        	(
	        	C_PyVector_FromValues(cx, cy, cz),
		        C_PyVector_New(), 0,
        		C_PyVector_FromValues(dx, dy, dz)
	        ), NULL);
	
	C_PyContainer_Append((PyObject *)container, unitBox);
	return (PyObject *)container;
}

PyObject *C_TograUnitCylinder(int slices) {

        int i;
        PyObject *top = C_PyVectorList_OfLength(slices + 2);
        PyObject *topn = C_PyVectorList_OfLength(slices + 2);
        PyObject *bottom = C_PyVectorList_OfLength(slices + 2);
        PyObject *bottomn = C_PyVectorList_OfLength(slices + 2);
        PyObject *sides = C_PyVectorList_OfLength(2 * (slices + 1));
        PyObject *sidesn = C_PyVectorList_OfLength(2 * (slices + 1));
        PyObject *container = C_PyContainer_OfLength(3);

        VL_APPEND(top, C_PyVector_FromValues(0, 0.5, 0));
        VL_APPEND(bottom, C_PyVector_FromValues(0, -0.5, 0));
        VL_APPEND(topn, C_PyVector_FromValues(0, 1, 0));
        VL_APPEND(bottomn, C_PyVector_FromValues(0, -1, 0));

        for(i = 0; i <= slices; i++) {
                VL_APPEND(top, C_PyVector_FromValues(
                                0.5*sin(i*2*M_PI/slices), 0.5,
                                0.5*cos(i*2*M_PI/slices)));
                VL_APPEND(topn, C_PyVector_FromValues(0, 1, 0));
                VL_APPEND(bottom, C_PyVector_FromValues(
                                0.5*sin((slices - i)*2*M_PI/slices), -0.5,
                                0.5*cos((slices - i)*2*M_PI/slices)));
                VL_APPEND(bottomn, C_PyVector_FromValues(0, -1, 0));
                VL_APPEND(sides, C_PyVector_FromValues(
                                0.5*sin(i*2*M_PI/slices), 0.5,
                                0.5*cos(i*2*M_PI/slices)));
                VL_APPEND(sides, C_PyVector_FromValues(
                                0.5*sin(i*2*M_PI/slices), -0.5,
                                0.5*cos(i*2*M_PI/slices)));
                VL_APPEND(sidesn, C_PyVector_FromValues(
                                0.5*sin(i*2*M_PI/slices), 0,
                                0.5*cos(i*2*M_PI/slices)));
                VL_APPEND(sidesn, C_PyVector_FromValues(
                                0.5*sin(i*2*M_PI/slices), 0,
                                0.5*cos(i*2*M_PI/slices)));
        }
        
        C_PyContainer_Append(container, 
                C_PyStaticRender_New(top, GL_TRIANGLE_FAN, topn, NULL, NULL, 
                                0, slices + 2, 0, 0));
        
        C_PyContainer_Append(container, 
                C_PyStaticRender_New(sides, GL_QUAD_STRIP, sidesn, NULL, NULL, 
                                0, 2*slices + 2, 0, 0));

        C_PyContainer_Append(container, 
                C_PyStaticRender_New(bottom, GL_TRIANGLE_FAN, bottomn, NULL, 
                                NULL, 0, slices + 2, 0, 0));
        
        return container;

}

PyObject *C_TograCylinder(double cx, double cy, double cz, double r, double h,
                int slices) {

        PyContainer *cylinder = (PyContainer *)C_TograUnitCylinder(slices);
        C_PyTransformContainer_Append(cylinder->transform, C_PyTransform_New(
                        C_PyVector_FromValues(cx, cy, cz),
                        C_PyVector_New(), 0,
                        C_PyVector_FromValues(r, h, r)), NULL);

        return (PyObject *)cylinder;

}
                               
PyObject *TograSphere(PyObject *self, PyObject *args) {

	double cx, cy, cz, r;
	int slices, segs;
	
	if (!PyArg_ParseTuple(args, "ddddii", &cx, &cy, &cz, &r, &slices, &segs)) {
		return NULL;
	}
	
	return C_TograSphere(cx, cy, cz, r, slices, segs);
	
}

PyObject *TograBox(PyObject *self, PyObject *args) {

	double cx, cy, cz, dx, dy, dz;
	
	if (!PyArg_ParseTuple(args, "dddddd", &cx, &cy, &cz, &dx, &dy, &dz)) {
		return NULL;
	}
		
	return C_TograBox(cx, cy, cz, dx, dy, dz);
	
}

PyObject *TograCylinder(PyObject *self, PyObject *args) {

        double cx, cy, cz, r, h;
        int slices;

        if (!PyArg_ParseTuple(args, "dddddi", &cx, &cy, &cz, &r, &h, &slices)) {
                return NULL;
        }

        return C_TograCylinder(cx, cy,  cz, r, h, slices);

}

PyObject *C_TograCircle(double cx, double cy, double r, int steps) {

        PyObject *circle = C_PyVectorList_OfLength(steps);
        int i;

        for (i = 0; i < steps; i++)
        {
                VL_APPEND(circle, C_PyVector_FromValues(
                                        cx + r * sin(i * 2 * M_PI / steps),
                                        cy + r * cos(i * 2 * M_PI / steps),
                                        0
                        ));
        }

        return circle;

}

PyObject *TograCircle(PyObject *self, PyObject *args) {

        double cx, cy, r;
        int steps;

        if (!PyArg_ParseTuple(args, "dddi", &cx, &cy, &r, &steps)) {
                return NULL;
        }

        return C_TograCircle(cx, cy, r, steps);

}

PyObject *C_TograExtrude(PyObject *vector_list, PyObject *extrude_path) {

        PyVectorList *vl = (PyVectorList *)vector_list;
        PyVector *v = (PyVector *)extrude_path;
        int i;
        double *p;
        PyVectorList *object_data;
        PyVector **q;

        object_data 
               = (PyVectorList *)C_PyVectorList_OfLength(vl->length * 2 + 2);


        p = object_data->values;
        
        for (i = 0; i < vl->length; i++) {
                *p++ = vl->values[i*3];
                *p++ = vl->values[i*3 + 1];
                *p++ = vl->values[i*3 + 2];
                *p++ = vl->values[i*3] + v->vector[0];
                *p++ = vl->values[i*3 + 1] + v->vector[1];
                *p++ = vl->values[i*3 + 2] + v->vector[2];
        }

        *p++ = vl->values[0];
        *p++ = vl->values[1];
        *p++ = vl->values[2];
        *p++ = vl->values[0] + v->vector[0];
        *p++ = vl->values[1] + v->vector[1];
        *p++ = vl->values[2] + v->vector[2];
        
        object_data->length = vl->length * 2 + 2;

        q = (PyVector **)object_data->pyVectors;
        p = object_data->values;
        for (i = 0; i < object_data->length; i++)
        {
                (*q++) = (PyVector *)C_PyVector_FromVector(p);
                p += 3;
        }

        return (PyObject *)object_data;
                        
}

PyObject *TograExtrude(PyObject *self, PyObject *args) {

        PyObject *vl, *v;
        
        if (!PyArg_ParseTuple(args, "OO", &vl, &v)) {
                return NULL;
        }

        return C_TograExtrude(vl, v);
        
}
