/*
 * Copyright 2003, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

staticforward PyTypeObject PyWobbleBehaviourType;

void C_PyWobbleBehaviour_Run(PyObject *self, double *v, double *n, int width,
		int height, int time);
static IDSBehaviour PyWobbleBehaviour_IDSBehaviour 
	= {C_PyWobbleBehaviour_Run};

static IC PyWobbleBehaviour_Interfaces = {NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, &PyWobbleBehaviour_IDSBehaviour};

PyObject *C_PyWobbleBehaviour_New(double xper, double yper, 
		double xmag, double ymag);

PyObject *PyWobbleBehaviour_New(PyObject *self, PyObject *args) {

	double xper = 1, yper = 1, xmag = 0.5, ymag = 0.5;
	
	if (!PyArg_ParseTuple(args, "|dddd", &xper, &yper, &xmag, &ymag)) {
		return NULL;
	}

	return C_PyWobbleBehaviour_New(xper, yper, xmag, ymag);
	
}

void C_PyWobbleBehaviour_Run(PyObject *self, double *v, double *n, int width,
		int height, int time) {

	int i, j;
	double *pos;
	PyWobbleBehaviour *d = (PyWobbleBehaviour *)self;
	double xper = d->xper + sin(2*time*3.142/d->xperper/1000)*d->xpermag;
	double yper = d->yper + sin(2*time*3.142/d->yperper/1000)*d->ypermag;
	
	pos = n;

	for (i = 0; i < width; i++)
	{
		double xdiff;
		xdiff = (d->xmag) * sin(2*i*3.142/xper/(width-1));
		for (j = 0; j < height; j++)
		{
			double ydiff;
			ydiff = d->ymag * sin(2*j*3.142/yper/(height-1));
			pos[0] += xdiff;
			pos[1] += ydiff;
			pos += 3;
		}
	}
}

PyObject *C_PyWobbleBehaviour_New(double xper, double yper, double xmag, 
		double ymag) {
	
	PyWobbleBehaviour *d = PyObject_New(PyWobbleBehaviour,
			&PyWobbleBehaviourType);


	d->interfaces = &PyWobbleBehaviour_Interfaces;
	
	d->xper = xper;
	d->yper = yper;
	d->xmag = xmag;
	d->ymag = ymag;
	d->xperper = 1;
	d->yperper = 1;
	d->ypermag = 0;
	d->ypermag = 0;
	
	return (PyObject *)d;
	
	/* (for now) */
	//C_PyDynamicSurface_FlatSurface(self, x0, y0, x1, y1);

}

static int PyWobbleBehaviour_Setattr(PyObject *obj, char *name,
		PyObject *val) {

	PyWobbleBehaviour *data = (PyWobbleBehaviour *)obj;
	
	if (strcmp(name, "xPeriodPeriod") == 0)
	{
		data->xperper = PyFloat_AsDouble(val);
	}
	else if (strcmp(name, "yPeriodPeriod") == 0)
	{
		data->yperper = PyFloat_AsDouble(val);
	}
	else if (strcmp(name, "xPeriodMagnitude") == 0)
	{
		data->xpermag = PyFloat_AsDouble(val);
	}
	else if (strcmp(name, "yPeriodMagnitude") == 0)
	{
		data->ypermag = PyFloat_AsDouble(val);
	}
	else
	{
		return -1;
	}

	return 0;
}	

static void PyWobbleBehaviour_Free(PyObject *this) {

}

static PyTypeObject PyWobbleBehaviourType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyWobbleBehaviour",
        sizeof(PyWobbleBehaviour),
        0,
        PyWobbleBehaviour_Free,
        0, /* print */
        0, /* getattr */
        PyWobbleBehaviour_Setattr, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
