#include "TograIncludes.h"

Partition_3D *create_partition_3d() {
	void *p = malloc(sizeof(Partition_3D));
	memset(p, 0, sizeof(Partition_3D));
	return (Partition_3D *)p;
}

#define SHIFT				\
{					\
	int _;				\
	for (_ = 0; _ < shift; _++)	\
		printf(" ");		\
}

void _partition_3d_dump(Partition_3D *p, int x, int y, int z, int shift) {

	int i;
	
	for (i = 0; i < PARTITION_ELTS; i++) {
		int tx, ty, tz;
                tx = (i >> (2 * BITS_PER_NODE)) & (BP1 - 1);
                tx <<= shift;
                tx |= x;

                ty = (i >> BITS_PER_NODE) & (BP1 - 1);
                ty <<= shift;
                ty |= y;

                tz = i & (BP1 - 1);
                tz <<= shift;
                tz += z;

		if (p->this_level[i] == NULL) {
			//printf("No element at (%d,%d,%d)\n", tx, ty, tz);
		} else {
			SHIFT
			printf("%p at (%d,%d,%d)\n", p->this_level[i], tx, ty, 
					tz);
		}
	}

	for (i = 0; i < PARTITION_ELTS; i++) {
		int tx, ty, tz;
                tx = (i >> (2 * BITS_PER_NODE)) & (BP1 - 1);
                tx <<= shift;
                tx |= x;

                ty = (i >> BITS_PER_NODE) & (BP1 - 1);
                ty <<= shift;
                ty |= y;

                tz = i & (BP1 - 1);
                tz <<= shift;
                tz += z;

		if (p->next_level[i] == NULL) {
			//printf("No recursion at (%d,%d,%d)\n", tx, ty, tz);
			continue;
		}
		
		SHIFT
		if (shift < 2) {
			printf("Descending to (%d, %d, %d)\n", tx, ty, tz);
			_partition_3d_dump(p->next_level[i], 
					tx, ty, tz, shift + BITS_PER_NODE);
		}
		else {
			printf("Bottoming out\n");
		}
	}
}

void partition_3d_dump(Partition_3D *p) {
	_partition_3d_dump(p, 0, 0, 0, 0);
}

void partition_3d_fixkey(int *x, int *y, int *z) {

	if ((*x) < 0)
		(*x) = -(*x) * 2 - 1;
	else
		(*x) *= 2;
	if ((*y) < 0)
		(*y) = -(*y) * 2 - 1;
	else
		(*y) *= 2;
	if ((*z) < 0)
		(*z) = -(*z) * 2 - 1;
	else
		(*z) *= 2;
}

void partition_3d_unfixkey(int *x, int *y, int *z) {

	if ((*x) % 2)
		(*x) = -((*x) + 1) / 2;
	else
		(*x) /= 2;
	if ((*y) % 2)
		(*y) = -((*y) + 1) / 2;
	else
		(*y) /= 2;
	if ((*z) % 2)
		(*z) = -((*z) + 1) / 2;
	else
		(*z) /= 2;
}

int partition_3d_insert(Partition_3D *p, int x, int y, int z, void *data) {

	unsigned int key;

	partition_3d_fixkey(&x, &y, &z);
	
	while (1) {
		if (x < BP1 && y < BP1 && z < BP1) {
			key = (x << (2 * BITS_PER_NODE));
			key |= (y << BITS_PER_NODE) | z;
			
			if (p->this_level[key] == NULL) {
				p->this_level[key] = data;
				return 0;
			}
			return -1;
		} else {
			key = ((x & (BP1 - 1)) << (2 * BITS_PER_NODE));
			key |= ((y & (BP1 - 1)) << BITS_PER_NODE);
			key |= (z & (BP1 - 1));
		
			x >>= BITS_PER_NODE;
			y >>= BITS_PER_NODE;
			z >>= BITS_PER_NODE;
			
			if (p->next_level[key] == NULL) {
				p->next_level[key] = create_partition_3d();
			}
			
			p = p->next_level[key];
		}
	}
}

void *partition_3d_retrieve(Partition_3D *p, int x, int y, int z) {

	unsigned int key;
	
	partition_3d_fixkey(&x, &y, &z);
	
	while (1) {
		if (x < BP1 && y < BP1 && z < BP1) {
			key = (x << (2 * BITS_PER_NODE));
			key |= (y << BITS_PER_NODE) | z;
			
			return p->this_level[key];
		} else {
			key = ((x & (BP1 - 1)) << (2 * BITS_PER_NODE));
			key |= ((y & (BP1 - 1)) << BITS_PER_NODE);
			key |= (z & (BP1 - 1));
			
			x >>= BITS_PER_NODE;
			y >>= BITS_PER_NODE;
			z >>= BITS_PER_NODE;
			
			if (p->next_level[key] == NULL) {
				return NULL;
			}
		
			p = p->next_level[key];
		}
	}
}

void _partition_3d_foreach(Partition_3D *p, partition_3d_foreach_func func,
		void *passthrough, int x, int y, int z, int depth) {

	int i;
	
	for (i = 0; i < PARTITION_ELTS; i++) {

		void *data = p->this_level[i];
		int tx, ty, tz;
		int ttx, tty, ttz;
		int r;
		
		tx = (i >> (2 * BITS_PER_NODE)) & (BP1 - 1);
		tx <<= depth;
		tx |= x;

		ty = (i >> BITS_PER_NODE) & (BP1 - 1);
		ty <<= depth;
		ty |= y;

		tz = i & (BP1 - 1);
		tz <<= depth;
		tz |= z;
	
		if (data == NULL)
			goto foreach_stage_2;
	
		ttx = tx;
		tty = ty;
		ttz = tz;
		partition_3d_unfixkey(&ttx, &tty, &ttz);
		
		r = func(ttx, tty, ttz, data, passthrough);
		if (r == 1)
		{
			p->this_level[i] = NULL;
		}
		
foreach_stage_2:	
		if (p->next_level[i] == NULL)
			continue;

		_partition_3d_foreach(p->next_level[i], func, passthrough,
				tx, ty, tz, depth + BITS_PER_NODE);
	}
}

void partition_3d_foreach(Partition_3D *p, partition_3d_foreach_func func,
		void *passthrough) {
	_partition_3d_foreach(p, func, passthrough, 0, 0, 0, 0);
}

