/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"
#include "PyInterpolatedPath_common.h"

#define IP_INITED	(1<<0)
#define IP_CALLED 	(1<<1)

#define FRAC(c,a,b) (((float)((c) - (a)))/((b)-(a)))

typedef struct {
        int curIndex;
        int is_called;
        int interpPointA;
        int interpPointB;
        int delay;
        int doCallback;
        PyObject *userPointer;
        int nextInterval;
        int intervalSize;
        int genIntervals;
        int wrap;
	int curTime;
} PyIPStatus;

int C_PyInterpolatedPath_GenerateTransform(PyObject *self, 
                PyObject *status, int curPos, int gen_forward, int gen_reverse,
                int apply) {
        PyInterpolatedPath *ip = (PyInterpolatedPath *)self;
        PyIPStatus *s = (PyIPStatus *)status;
	PyObject *translate1, *translate2;
	int pointIndex1, pointIndex2;

        if (!(ip->isInited)) {
                ip->baseTime = curPos;
                ip->isInited = 1;
        }
        
	ip->lastStatus = status;

        if (curPos - ip->baseTime - s->delay >= ip->maxTime && 
                        s->doCallback != 0) {

                translate1 = ip->pathPoints[ip->pathPointsSize - 1];
		if (gen_forward && apply)
			IINTERPOLATEDPATHSEGMENT_TRANSFORMBYME(ip->interpolator,
					translate1, translate1, 0);
		else if (gen_reverse && apply)
			IINTERPOLATEDPATHSEGMENT_INVERTBYME(ip->interpolator,
					translate1, translate1, 0);
				
                if (!(s->is_called)) {
                        s->is_called = 1;
			Py_INCREF(self);
			if (s->userPointer == NULL) {
				Py_INCREF(Py_None);
				AppendMessage(std_queue, 
					MESSAGE_INTERPOLATED_PATH_COMPLETE, 
					self, Py_None);
			} else {
				Py_INCREF(s->userPointer);
				AppendMessage(std_queue, 
					MESSAGE_INTERPOLATED_PATH_COMPLETE, 
					self, s->userPointer);
			}
                }
                return 1;
        } else {
                s->is_called = 0;
        }

	
        curPos -= ip->baseTime + s->delay;
	s->curTime = curPos;
        if (curPos > 0) {
		if (curPos - (curPos % ip->maxTime) > 0)
			s->delay += ip->maxTime;
                curPos %= ip->maxTime;
		s->curTime = curPos;
        } else {
		s->curTime = curPos;
                translate1 = ip->pathPoints[0];
		if (gen_forward && apply)
			IINTERPOLATEDPATHSEGMENT_TRANSFORMBYME(ip->interpolator,
					translate1, translate1, 0);
		else if (gen_reverse && apply)
			IINTERPOLATEDPATHSEGMENT_INVERTBYME(ip->interpolator,
					translate1, translate1, 0);
                return 1;
        }

	//printf("%d %d\n", curPos, s->nextInterval);
	if (s->genIntervals == 0)
		goto notest;
	if (curPos >= s->nextInterval)
	{
		if (curPos + s->intervalSize > ip->maxTime && s->wrap == 1)
			goto notest;
	
		s->wrap = 0;
domessage:	
		Py_INCREF(self);
		Py_INCREF(Py_None);
		AppendMessage(std_queue, MESSAGE_INTERPOLATED_PATH_INTERVAL, 
				self, Py_None);

		s->nextInterval += s->intervalSize;
		if (s->nextInterval > ip->maxTime) {
			s->nextInterval -= ip->maxTime;
			s->wrap = 1;
			goto notest;
		}
	} else if (curPos < s->intervalSize && s->wrap == 0 &&
			s->nextInterval + s->intervalSize > ip->maxTime) {

		goto domessage;

	}
		
notest:

	if (curPos < s->interpPointA) {
		s->interpPointB = ip->pathIndexes[0];
                s->curIndex = -1;
                s->interpPointA = 0;
	}

	while (curPos >= s->interpPointB) {
		s->interpPointA = s->interpPointB;
		if (++(s->curIndex) == ip->pathPointsSize - 1) {
			s->interpPointB = ip->maxTime;
                } else {
			s->interpPointB = ip->pathIndexes[s->curIndex + 1];
		}
	}
	
	if (s->curIndex == -1) {
		pointIndex1 = pointIndex2 = 0;
	} else if (s->curIndex == ip->pathPointsSize - 1) {
		pointIndex1 = pointIndex2 = s->curIndex;
	} else {
		pointIndex1 = s->curIndex;
		pointIndex2 = pointIndex1 + 1;
	}

	translate2 = ip->pathPoints[pointIndex2];
	translate1 = ip->pathPoints[pointIndex1];

	if (gen_forward && apply)
		IINTERPOLATEDPATHSEGMENT_TRANSFORMBYME(ip->interpolator,
				translate1, translate2, 
				FRAC(curPos, s->interpPointA, s->interpPointB));
	else if (gen_reverse && apply)
		IINTERPOLATEDPATHSEGMENT_INVERTBYME(ip->interpolator,
				translate1, translate2, 
				FRAC(curPos, s->interpPointA, s->interpPointB));
	

	return 1;
}

int C_PyInterpolatedPath_TransformByMe(PyObject *this) {
	PyInterpolatedPath *ip = (PyInterpolatedPath *)this;
	PyIPStatus *s = (PyIPStatus *)ip->lastStatus;

	int pointIndex1, pointIndex2;
	PyObject *translate1, *translate2;
	
	if (s->curIndex == -1) {
		pointIndex1 = pointIndex2 = 0;
	} else if (s->curIndex == ip->pathPointsSize - 1) {
		pointIndex1 = pointIndex2 = s->curIndex;
	} else {
		pointIndex1 = s->curIndex;
		pointIndex2 = pointIndex1 + 1;
	}

	translate2 = ip->pathPoints[pointIndex2];
	translate1 = ip->pathPoints[pointIndex1];
	
	IINTERPOLATEDPATHSEGMENT_TRANSFORMBYME(ip->interpolator, translate1, 
		translate2, FRAC(s->curTime, s->interpPointA, s->interpPointB));

	return 1;
}

int C_PyInterpolatedPath_InvertByMe(PyObject *this) {
	PyInterpolatedPath *ip = (PyInterpolatedPath *)this;
	PyIPStatus *s = (PyIPStatus *)ip->lastStatus;
	
	int pointIndex1, pointIndex2;
	PyObject *translate1, *translate2;

	if (s->curIndex == -1) {
		pointIndex1 = pointIndex2 = 0;
	} else if (s->curIndex == ip->pathPointsSize - 1) {
		pointIndex1 = pointIndex2 = s->curIndex;
	} else {
		pointIndex1 = s->curIndex;
		pointIndex2 = pointIndex1 + 1;
	}

	translate2 = ip->pathPoints[pointIndex2];
	translate1 = ip->pathPoints[pointIndex1];
	
	IINTERPOLATEDPATHSEGMENT_INVERTBYME(ip->interpolator, translate1, 
		translate2, FRAC(s->curTime, s->interpPointA, s->interpPointB));
	
	return 1;
}

int C_PyInterpolatedPath_RegisterSize(PyObject *self, int *result) {
        *result = sizeof(PyIPStatus);
	return 1;
}

int C_PyInterpolatedPath_setNotificationInterval(PyInterpolatedPath *self,
		int handle, int offset, int interval) {

	PyIPStatus *s = (PyIPStatus *)handle;
	s->nextInterval = offset;
	s->intervalSize = interval;
	s->genIntervals = 1;

	return 1;
}

int C_PyInterpolatedPath_setIterationScheme(PyInterpolatedPath *self,
		int handle, int scheme) {

	PyIPStatus *s = (PyIPStatus *)handle;
	
	if (scheme == ITERATION)
	{
		s->doCallback = 0;
	}
	else
	{
		s->doCallback = 1;
		if (s->userPointer == NULL)
			s->is_called = 1;
		else
			s->is_called = 0;
	}

	return 1;
}

int C_PyInterpolatedPath_setDelay(PyInterpolatedPath *self, int handle,
		int delay)
{
	PyIPStatus *s = (PyIPStatus *)handle;
	s->delay = delay;

	return 1;
}

int C_PyInterpolatedPath_setCallback(PyInterpolatedPath *self, int handle,
		PyObject *callback)
{

	PyIPStatus *s = (PyIPStatus *)handle;

	if (callback != Py_None) {
		Py_INCREF(callback);
	}

	if (s->userPointer != NULL) {
		Py_DECREF(s->userPointer);
	}

	if (callback != Py_None) {
		s->userPointer = callback;
		s->doCallback = 1;
		s->is_called = 0;
	} else {
		s->userPointer = NULL;
		s->doCallback = 0;
	}

	return 1;
}	

int C_PyInterpolatedPath_Register(PyObject *self, PyObject *status) {

	PyInterpolatedPath *ip = (PyInterpolatedPath *)self;
        PyIPStatus *s = (PyIPStatus *)status;
        
	s->doCallback = 0;
	s->userPointer = NULL;
        s->delay = 0;

	s->curIndex = 0;
        s->is_called = 0;
        s->interpPointA = 0;
	s->nextInterval = ip->maxTime + 1;
	s->genIntervals = 0;
	s->wrap = 0;
	
	s->interpPointB = ip->pathIndexes[1];

	return 1;

}

int C_PyInterpolatedPath_DeRegister(PyObject *self, PyObject *status)
{

        PyIPStatus *s = (PyIPStatus *)status;

        if (s->userPointer != NULL) {
                Py_DECREF(s->userPointer);
        }

	return 1;
}
