/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"
#include "PyLogo_common.h"

#define VERTICAL_BYPARAM	0
#define TOP			1
#define VERTICAL_CENTER		2
#define BOTTOM			3
#define HORIZONTAL_BYPARAM	0
#define LEFT			4
#define HORIZONTAL_CENTER	8
#define RIGHT			12

void C_PyLogo_SetupSubTexture(PyLogo *object, int *bits, int width, int height) 
{

	int w = 2, h = 2;
	int *lnew;
	int i;
	
	while (w < width)
		w *= 2;
	
	while (h < height)
		h *= 2;

	object->xfrac = (float)width/w;
	object->yfrac = (float)height/h;
	
	lnew = malloc(w* h * 4);
	
	for (i = 0; i < height; i++)
		memcpy(lnew + i * w, bits + i * width, width * 4);
	
	object->logo = C_PyStaticTexture_New(lnew, w, h, GL_RGBA);

	free(lnew);
	
}

int getScreenWidth(void);
int getScreenHeight(void);

#define GET_YPOS(x)	(x & 0x3)
#define GET_XPOS(x)	(x & 0xC)

int C_PyLogo_Render(PyObject *self, int time, int picking) {

        PyLogo *r = (PyLogo *)self;

	int width = getScreenWidth();
	int height = getScreenHeight();
	int xpos, ypos;
	float col[] = {1.0,1.0,1.0,1.0};

	if (GET_YPOS(r->position) == TOP)
		ypos = 0;
	else if (GET_YPOS(r->position) == VERTICAL_CENTER)
		ypos = (height - r->height) / 2;
	else if (GET_YPOS(r->position) == BOTTOM)
		ypos = height - r->height;
	else
		ypos = r->y;

	if (GET_XPOS(r->position) == LEFT)
		xpos = 0;
	else if (GET_XPOS(r->position) == HORIZONTAL_CENTER)
		xpos = (width - r->width) / 2;
	else if (GET_XPOS(r->position) == RIGHT)
		xpos = width - r->width;
	else
		xpos = r->x;
	
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	
	gluOrtho2D(0.0, width, height, 0.0);

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, col);
	ITEXTURE_SETTEXTURE(r->logo, time);
	
	glBegin(GL_QUADS);
	glNormal3f(0,0,1);
	glTexCoord2f(0.0, 0.0);
	glVertex3f(xpos, ypos, 0);
	glTexCoord2f(0.0, r->yfrac);
	glVertex3f(xpos, ypos + r->height, 0);
	glTexCoord2f(r->xfrac, r->yfrac);
	glVertex3f(xpos + r->width, ypos + r->height, 0);
	glTexCoord2f(r->xfrac, 0.0);
	glVertex3f(xpos + r->width, ypos, 0);
	glEnd();
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	return 1;
}

int C_PyLogo_Compile(PyObject *self, int *result) {
	*result = FALSE;
	return 1;
}
