#include "TograIncludes.h"
#include "PyRenderControl_common.h"

PyMaterial *current_colour;

void C_PyRenderControl_InternalStep(void);

struct _C_PyRenderControl_LightObj {
        int active;
        float position[4];
        float diffuse[4];
        float specular[4];
        PyObject *transform;
};

#define NUM_LIGHTS      8

inline void _C_PyRenderControl_InitialiseLighting(PyRenderControl *rc) {

        int i;
        
        rc->lights = malloc(NUM_LIGHTS * sizeof(_C_PyRenderControl_LightObj));

        for (i = 0; i < NUM_LIGHTS; i++)
        {
                memset(&rc->lights[i], 0, sizeof(_C_PyRenderControl_LightObj));
                rc->lights[i].active = LIGHT_UNUSED;
                rc->lights[i].transform = NULL;
        }
        
}

inline void _C_PyRenderControl_DoLighting(PyRenderControl *rc, int time) {

        int i;

        for (i = 0; i < NUM_LIGHTS; i++) {
                if (rc->lights[i].active == LIGHT_ACTIVE) {
                        glPushMatrix();
                        if (rc->lights[i].transform != NULL) {
                                ITRANSFORM_GENERATETRANSFORM(
					rc->lights[i].transform, NULL, time,
                                        1, 0, 1);
                        }
                        glLightfv(GL_LIGHT0 + i, GL_POSITION, 
                                        rc->lights[i].position);
                        glLightfv(GL_LIGHT0 + i, GL_DIFFUSE, 
                                        rc->lights[i].diffuse);
                        glLightfv(GL_LIGHT0 + i, GL_SPECULAR, 
                                        rc->lights[i].specular);
			glPopMatrix();
		}
        }

}

MsgQueue	*std_queue	= NULL;

static PyRenderControl *this_ptr;

int getScreenWidth()
{
	return glutGet(GLUT_WINDOW_WIDTH);
}

int getScreenHeight()
{
	return glutGet(GLUT_WINDOW_HEIGHT);
}

int C_PyRenderControl_deleteLight(PyRenderControl *object, int light) {
	if (object->lights[light].active == LIGHT_UNUSED)
		return 0;

	object->lights[light].active = LIGHT_UNUSED;
	glDisable(GL_LIGHT0 + light);

	return 1;
}

int C_PyRenderControl_newLight(PyRenderControl *object, int *return_value) { 
	int i;
	for (i = 0; i < NUM_LIGHTS; i++) {
		if (object->lights[i].active == LIGHT_UNUSED)
			break;
	}

	if (i == NUM_LIGHTS)
		return 0;

	object->lights[i].active = LIGHT_ACTIVE;
	object->lights[i].transform = NULL;
	
	glEnable(GL_LIGHT0 + i);
	*return_value = i;

	return 1;

}

int C_PyRenderControl_setLightPosition(PyRenderControl *object, int light,
		PyObject *position) {

	int i;
	
	PyVector *p = (PyVector *)position;
	/*
	 * Iteration necessary for conversion between double and float
	 */
	for (i = 0; i < 3; i++)
		object->lights[light].position[i] = p->vector[i];

	return 1;
	
}

int C_PyRenderControl_setLightDiffuse(PyRenderControl *object, int light,
		PyObject *diffuse) {

	PyColour *c = (PyColour *)diffuse;
	memcpy(object->lights[light].diffuse, c->colour, 4 * sizeof(float));

	return 1;
}

int C_PyRenderControl_setLightSpecular(PyRenderControl *object, int light,
		PyObject *specular) {
	
	PyColour *c = (PyColour *)specular;
	memcpy(object->lights[light].specular, c->colour, 4 *sizeof(float));

	return 1;

}

int C_PyRenderControl_setLightTransform(PyRenderControl *object, int light,
		PyObject *transform) {

	int r;
	
	Py_XDECREF(object->lights[light].transform);

	Py_XINCREF(transform);

	/*
	 * for now
	 */
	ITRANSFORM_REGISTERSIZE(transform, &r);
	assert(r == 0);
	
	object->lights[light].transform = transform;

	return 1;
}

void C_PyRenderControl_Init(PyRenderControl *object, int isFullScreen) {
	
	int args = 2;
	char *argv[] = {"python", "derivative.py", NULL};
        float lm[4] = {0.5, 0.5, 0.5, 1};
	
	this_ptr = object;

	glutInit(&args, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STENCIL);
	glutInitWindowSize(800, 800);
	glutInitWindowPosition(300, 0);
	glutCreateWindow("Welcome to the World of Togra!");
	
	if (isFullScreen)
		glutFullScreen();

        glClearColor(1.0, 1.0, 1.0, 0.0);
        glColor3f(0, 0, 0);
        glPointSize(4);
        glViewport(0, 0, 480, 480);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluPerspective(60, 1, 1.5, 1000);

        glEnable(GL_DEPTH_TEST);
        glEnable(GL_LIGHTING);
        glEnable(GL_NORMALIZE);
        glEnable(GL_CULL_FACE);
        //glEnable(GL_POLYGON_SMOOTH);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glMatrixMode(GL_MODELVIEW);

        glLoadIdentity();

        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lm);
        _C_PyRenderControl_InitialiseLighting(rc);
}

void C_PyRenderControl_InternalStep(void)
{

        double *from, *to, *up;
        long time;

        PyRenderControl *rc = (PyRenderControl *)this_ptr;
        
	from = ((PyVector *)(rc->viewFrom))->vector;
        to = ((PyVector *)(rc->viewTo))->vector;
        up = ((PyVector *)(rc->viewUp))->vector;

	
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        gluLookAt(from[0], from[1], from[2], to[0], to[1], to[2], up[0],
                        up[1], up[2]);

        time = GetTime() - rc->baseTime - rc->pausedAmount;
	rc->lastTime = time;

        if (rc->isPaused) {
		_C_PyRenderControl_DoLighting(rc, rc->pausedTime - 
				rc->baseTime - rc->pausedAmount);
                IRENDER_RENDER(rc->baseContainer, 
				rc->pausedTime - rc->baseTime - 
				rc->pausedAmount, 0);
        } else {
		_C_PyRenderControl_DoLighting(rc, time);
                IRENDER_RENDER(rc->baseContainer, time, 0);
        }
	
	/* TODO: Add this back in when Mixins work again
	C_StepMixins(rc->mixins, time, rc->isPaused, rc->pausedAmount);
	*/
	
        Py_BEGIN_ALLOW_THREADS
        glutSwapBuffers();
        glutPostRedisplay();
        Py_END_ALLOW_THREADS
	SendMessages(std_queue, rc->messageFunction);	
}

int C_PyRenderControl_run(PyRenderControl *rc) {

        PyMaterial *default_colour = (PyMaterial *)C_PyMaterial_New(
			C_PyColour_New(0.7, 0.7, 0.7, 1), 
			C_PyColour_New(0, 0, 0, 1),
			C_PyColour_New(0, 0, 0, 1),
			1);
	current_colour = default_colour;
	
	int r;

	rc->baseTime = GetTime();
	rc->lastTime = rc->baseTime;

        IRENDER_COMPILE(rc->baseContainer, &r);

	std_queue = NewQueue();

        glutDisplayFunc(C_PyRenderControl_InternalStep);
        glutMainLoop();
	return 1;
}

void C_PyRenderControl_KBFunc(unsigned char key, int x, int y)
{

        PyRenderControl *rc = (PyRenderControl *)this_ptr;

        PyObject *tup = PyTuple_New(3);
        PyTuple_SetItem(tup, 0, PyInt_FromLong(key));
        PyTuple_SetItem(tup, 1, PyInt_FromLong(x));
        PyTuple_SetItem(tup, 2, PyInt_FromLong(y));

        PyObject_CallObject(rc->keyFunction, tup);

        return;
}

int C_PyRenderControl_stop(PyRenderControl *rc) {

        glutIdleFunc(NULL);
        rc->isPaused = TRUE;
	rc->pausedTime = GetTime();

	return 1;
}

int C_PyRenderControl_pause(PyRenderControl *rc) {

	rc->isPaused = TRUE;
	rc->pausedTime = GetTime();

	return 1;
}

int C_PyRenderControl_unPause(PyRenderControl *rc) {

	rc->isPaused = FALSE;
	rc->pausedAmount += GetTime() - rc->pausedTime;

	return 1;
}
