#include "TograIncludes.h"
#include "PyRenderControl_common.h"

PyMaterial *current_colour;

gboolean C_PyRenderControl_InternalStep(PyRenderControl *rc);
GtkWidget 	*gl_area;
GtkWidget 	*window;

struct _C_PyRenderControl_LightObj {
        int active;
        float position[4];
        float diffuse[4];
        float specular[4];
        PyObject *transform;
};

gboolean C_PyRenderControl_MakeCurrent()
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (gl_area);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (gl_area);

  return gdk_gl_drawable_make_current (gldrawable, glcontext);
}

#define NUM_LIGHTS      8

inline void _C_PyRenderControl_InitialiseLighting(PyRenderControl *rc) {

        int i;
        
        rc->lights = malloc(NUM_LIGHTS * sizeof(_C_PyRenderControl_LightObj));

        for (i = 0; i < NUM_LIGHTS; i++)
        {
                memset(&rc->lights[i], 0, sizeof(_C_PyRenderControl_LightObj));
                rc->lights[i].active = LIGHT_UNUSED;
                rc->lights[i].transform = NULL;
        }
        
}

inline void _C_PyRenderControl_DoLighting(PyRenderControl *rc, int time) {

        int i;

        for (i = 0; i < NUM_LIGHTS; i++) {
                if (rc->lights[i].active == LIGHT_ACTIVE) {
                        glPushMatrix();
                        if (rc->lights[i].transform != NULL) {
                                ITRANSFORM_GENERATETRANSFORM(
					rc->lights[i].transform, NULL, time,
                                        1, 0, 1);
                        }
                        glLightfv(GL_LIGHT0 + i, GL_POSITION, 
                                        rc->lights[i].position);
                        glLightfv(GL_LIGHT0 + i, GL_DIFFUSE, 
                                        rc->lights[i].diffuse);
                        glLightfv(GL_LIGHT0 + i, GL_SPECULAR, 
                                        rc->lights[i].specular);
			glPopMatrix();
		}
        }

}

MsgQueue	*std_queue	= NULL;

int getScreenWidth(void)
{
	return gl_area->allocation.width;
}
int getScreenHeight(void)
{
	return gl_area->allocation.height;
}

static gboolean _delete_event(GtkWidget *widget, GdkEvent *event, gpointer data){
        return FALSE;
}

static void _destroy(GtkWidget *widget, gpointer data)
{
        gtk_main_quit();
	exit(0);
}

gint _reshape(GtkWidget *widget, GdkEventConfigure *event, gpointer data)
{

        int w = widget->allocation.width;
        int h = widget->allocation.height;

        if (C_PyRenderControl_MakeCurrent())
        {
            glViewport (0, 0, w, h);
        }

        return 0;
}

gint _init_gl_area(GtkWidget *widget, gpointer data);

int C_PyRenderControl_deleteLight(PyRenderControl *object, int light) {
	if (object->lights[light].active == LIGHT_UNUSED)
		return 0;

	object->lights[light].active = LIGHT_UNUSED;
        C_PyRenderControl_MakeCurrent();
	glDisable(GL_LIGHT0 + light);

	return 1;
}

int C_PyRenderControl_newLight(PyRenderControl *object, int *return_value) { 
	int i;
	for (i = 0; i < NUM_LIGHTS; i++) {
		if (object->lights[i].active == LIGHT_UNUSED)
			break;
	}

	if (i == NUM_LIGHTS)
		return 0;

	object->lights[i].active = LIGHT_ACTIVE;
	object->lights[i].transform = NULL;
	
        C_PyRenderControl_MakeCurrent();
	glEnable(GL_LIGHT0 + i);
	*return_value = i;

	return 1;

}

int C_PyRenderControl_setLightPosition(PyRenderControl *object, int light,
		PyObject *position) {

	int i;
	
	PyVector *p = (PyVector *)position;
	/*
	 * Iteration necessary for conversion between double and float
	 */
	for (i = 0; i < 3; i++)
		object->lights[light].position[i] = p->vector[i];

	return 1;
	
}

int C_PyRenderControl_setLightDiffuse(PyRenderControl *object, int light,
		PyObject *diffuse) {

	PyColour *c = (PyColour *)diffuse;
	memcpy(object->lights[light].diffuse, c->colour, 4 * sizeof(float));

	return 1;
}

int C_PyRenderControl_setLightSpecular(PyRenderControl *object, int light,
		PyObject *specular) {
	
	PyColour *c = (PyColour *)specular;
	memcpy(object->lights[light].specular, c->colour, 4 *sizeof(float));

	return 1;

}

int C_PyRenderControl_setLightTransform(PyRenderControl *object, int light,
		PyObject *transform) {

	int r;
	
	Py_XDECREF(object->lights[light].transform);

	Py_XINCREF(transform);

	/*
	 * for now
	 */
	ITRANSFORM_REGISTERSIZE(transform, &r);
	assert(r == 0);
	
	object->lights[light].transform = transform;

	return 1;
}

void C_PyRenderControl_Init(PyRenderControl *object, int isFullScreen) {
        GdkGLConfig *glconfig;

	gtk_init(NULL, NULL);
        gtk_gl_init (NULL, NULL);
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	if (isFullScreen) {
		gtk_window_fullscreen(GTK_WINDOW(window));
	} else {
		gtk_window_set_default_size(GTK_WINDOW(window), 800, 800);
	}

	g_signal_connect(G_OBJECT(window), "delete_event",
			G_CALLBACK(_delete_event), NULL);
	g_signal_connect(G_OBJECT(window), "destroy",
			G_CALLBACK(_destroy), NULL);
	gl_area = gtk_drawing_area_new ();

        glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGBA | 
                    GDK_GL_MODE_DEPTH | GDK_GL_MODE_DOUBLE);
        g_assert (glconfig != NULL);
        gtk_widget_set_gl_capability (gl_area, glconfig, NULL, TRUE, 
            GDK_GL_RGBA_TYPE);

	/*
	 * GtkWidget *widget, GdkEvent *event, gpointer user_data
	 * for delete_event
	 *
	 * GtkObject *object, gpointer user_data 
	 * for destroy
	 */

	gtk_widget_set_events(GTK_WIDGET(gl_area), 
			GDK_EXPOSURE_MASK | 
			GDK_BUTTON_PRESS_MASK |
			GDK_BUTTON_RELEASE_MASK |
			GDK_POINTER_MOTION_MASK |
			GDK_POINTER_MOTION_HINT_MASK |
			GDK_KEY_PRESS_MASK);

	gtk_container_add(GTK_CONTAINER(window), gl_area);

	g_signal_connect(GTK_OBJECT(gl_area), "realize",
			G_CALLBACK(_init_gl_area), object);
	g_signal_connect(GTK_OBJECT(gl_area), "configure_event",
			G_CALLBACK(_reshape), NULL);
	/*
	 * GtkWidget *widget, GdkEventConfigure *event, gpointer user_data
	 * for configure_event
	 *
	 * GtkWidget *widget, gpointer user_data
	 * for realize
	 */

	gtk_widget_show(window);
	gtk_widget_show(gl_area);
}

gint _init_gl_area(GtkWidget *widget, gpointer object)
{

        float lm[4] = {0.5, 0.5, 0.5, 1};
        PyRenderControl *rc = (PyRenderControl *)object;
        if (!C_PyRenderControl_MakeCurrent())
        {
                printf("Couldn't make GL area current on realize!\n");
                exit(1);
        }

        glClearColor(1.0, 1.0, 1.0, 0.0);
        glColor3f(0, 0, 0);
        glPointSize(4);
        glViewport(0, 0, 480, 480);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluPerspective(60, 1, 1.5, 1000);

        glEnable(GL_DEPTH_TEST);
        glEnable(GL_LIGHTING);
        glEnable(GL_NORMALIZE);
        glEnable(GL_CULL_FACE);
        //glEnable(GL_POLYGON_SMOOTH);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glMatrixMode(GL_MODELVIEW);

        glLoadIdentity();

        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lm);
        _C_PyRenderControl_InitialiseLighting(rc);

	g_idle_add((GtkFunction)C_PyRenderControl_InternalStep, object);

	return TRUE;
}

gboolean C_PyRenderControl_InternalStep(PyRenderControl *rc)
{

        double *from, *to, *up;
        long time;

        C_PyRenderControl_MakeCurrent();
        
	from = ((PyVector *)(rc->viewFrom))->vector;
        to = ((PyVector *)(rc->viewTo))->vector;
        up = ((PyVector *)(rc->viewUp))->vector;

	
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        gluLookAt(from[0], from[1], from[2], to[0], to[1], to[2], up[0],
                        up[1], up[2]);

        time = GetTime() - rc->baseTime - rc->pausedAmount;
	rc->lastTime = time;

        if (rc->isPaused) {
		_C_PyRenderControl_DoLighting(rc, rc->pausedTime - 
				rc->baseTime - rc->pausedAmount);
                IRENDER_RENDER(rc->baseContainer, 
				rc->pausedTime - rc->baseTime - 
				rc->pausedAmount, 0);
        } else {
		_C_PyRenderControl_DoLighting(rc, time);
                IRENDER_RENDER(rc->baseContainer, time, 0);
        }
	
	/* TODO: Add this back in when Mixins work again
	C_StepMixins(rc->mixins, time, rc->isPaused, rc->pausedAmount);
	*/
	
        Py_BEGIN_ALLOW_THREADS
        {
            GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (gl_area);
            if (gdk_gl_drawable_is_double_buffered (gldrawable))
              gdk_gl_drawable_swap_buffers (gldrawable);
            else
              glFlush();
        }
        Py_END_ALLOW_THREADS
	SendMessages(std_queue, rc->messageFunction);	

        return TRUE;
}

int C_PyRenderControl_run(PyRenderControl *rc) {

        PyMaterial *default_colour = (PyMaterial *)C_PyMaterial_New(
			C_PyColour_New(0.7, 0.7, 0.7, 1), 
			C_PyColour_New(0, 0, 0, 1),
			C_PyColour_New(0, 0, 0, 1),
			1);
	current_colour = default_colour;
	
	int r;

	rc->baseTime = GetTime();
	rc->lastTime = rc->baseTime;

        IRENDER_COMPILE(rc->baseContainer, &r);

	std_queue = NewQueue();

	gtk_main();
	return 1;
}

gboolean C_PyRenderControl_KBFunc(GtkWidget *widget, GdkEventKey *e,
                gpointer data)
{

        PyRenderControl *rc = (PyRenderControl *)data;

        PyObject *tup = PyTuple_New(2);
        PyTuple_SetItem(tup, 0, PyInt_FromLong(e->keyval));
        PyTuple_SetItem(tup, 1, PyInt_FromLong(e->state));

        PyObject_CallObject(rc->keyFunction, tup);

        return 0;
}

int C_PyRenderControl_stop(PyRenderControl *rc) {

	gtk_main_quit();
        rc->isPaused = TRUE;
	rc->pausedTime = GetTime();

	return 1;
}

int C_PyRenderControl_pause(PyRenderControl *rc) {

	rc->isPaused = TRUE;
	rc->pausedTime = GetTime();

	return 1;
}

int C_PyRenderControl_unPause(PyRenderControl *rc) {

	rc->isPaused = FALSE;
	rc->pausedAmount += GetTime() - rc->pausedTime;

	return 1;
}
