#include "TograIncludes.h"
#include "PyStaticRender_common.h"

void C_StaticRender_Lists_Static(PyObject *self, double *v, double *n, 
		int *faces, float *texture, int length, int type, 
		int normalMode, int tMode);
int C_PyStaticRender_Compile(PyObject *self, int *result);

int C_PyStaticRender_Render(PyObject *self, int time, int picking) {

        PyStaticRender *r = (PyStaticRender *)self;

        if (r->displayList == 0) {
		int result;
                if (!C_PyStaticRender_Compile(self, &result))
			return 0;
        }

        glCallList(r->displayList);
       
	return 1;
}

int C_PyStaticRender_Compile(PyObject *self, int *result) {

        double *v, *n = NULL;
        int *f = NULL;
        float *t = NULL;
        int length;
	
        PyStaticRender *r = (PyStaticRender *)self;

        if (r->displayList == 0)
        {
        
                IVECTORLIST_VALUES(r->vectors, &v);
                if (r->normals != NULL) {
                        IVECTORLIST_VALUES(r->normals, &n);
                }

                if (r->faces != NULL) {
                        IINTLIST_INTS(r->faces, &f);
			IINTLIST_LENGTH(r->faces, &length);
                }
		else
		{
			IVECTORLIST_LENGTH(r->vectors, &length);
		}
                
                r->displayList = glGenLists(1);

                if (r->textureCoords != NULL) {
			IFLOATLIST_FLOATS(r->textureCoords, &t);
                }
                
                if (r->displayList == 0)
                {
                        /* FIXME: Is this the correct behaviour? */
                        return 0;
                }
                glNewList(r->displayList, GL_COMPILE);

                C_StaticRender_Lists_Static(self, v, n, f, t, length, r->mode,
                                r->normalMode, r->textureMode);
                glEndList();
        }

	*result = 1;
        return 1;
                
}

inline void _texture(int mode, float *texture, int pos)
{

        if (texture == NULL)
                return;

        if (mode == 0) /* 2D Textures */
        {
                glTexCoord2f(texture[pos*2], texture[pos*2+1]);
        }
}

void C_StaticRender_Lists_Static(PyObject *self, double *v, double *n, 
		int *faces, float *texture, int length, int type, 
		int normalMode, int tMode) {

        int i;

        if (normalMode == 1)
        {

                int n_pos = 0, v_pos = 0;
                
                assert(n != NULL);
                

                switch (type) {
                        case GL_QUAD_STRIP:
                                
                                glBegin(GL_QUADS);

                                while (v_pos + 4 <= length) {

                                        glNormal3dv(&n[(n_pos++)*3]);
                                        _texture(tMode, texture, v_pos);
                                        glVertex3dv(&v[(v_pos)*3]);
                                        _texture(tMode, texture, v_pos + 1);
                                        glVertex3dv(&v[(v_pos+1)*3]);
                                        _texture(tMode, texture, v_pos + 3);
                                        glVertex3dv(&v[(v_pos+3)*3]);
                                        _texture(tMode, texture, v_pos + 2);
                                        glVertex3dv(&v[(v_pos+2)*3]);

                                        v_pos += 2;
                                }

                                glEnd();
                                
                                break;

                default:
                        assert(!"Per-face normals for this render mode not yet"
                                " implemented!");
                }


                return;
                
        }

        
        glBegin(type);
	
        if (faces == NULL) {
        
                for (i = 0; i < length; i++) {
                
                        if (n != NULL) {
                                glNormal3dv(&n[i*3]);
                        }

                        _texture(tMode, texture, i);
                        glVertex3dv(&v[i*3]);

                }
                
        } else {
                for (i = 0; i < length; i++) {
                        if (n != NULL) {
                                glNormal3dv(&n[faces[i] * 3]);
                        }
                        _texture(tMode, texture, faces[i]);
                        glVertex3dv(&v[faces[i] * 3]);
                }
        }


        glEnd();
                        
}

void C_PyStaticRender_QuadStrip_GenNormals(PyObject *vectors,
                PyObject **normals, PyObject **faces) {

        PyVectorList *v = (PyVectorList *)vectors;
        PyVectorList *n = (PyVectorList *)C_PyVectorList_New();
        int i;

        for (i = 0; i + 2 < v->valuesSize; i += 2)
        {
                PyObject *v0;
		PyObject *v1;
		PyObject *v2;
                (PyVector *)C_PyVector_Subtract(v->vectors[i + 1], 
						v->vectors[i], &v0);
	        (PyVector *)C_PyVector_Subtract(v->vectors[i + 3],
						v->vectors[i + 1], &v1);

		
		C_PyVector_Cross((PyObject *)v0, (PyObject *)v1, &v2);
                IVECTORLIST_APPEND((PyObject *)n, v2); 
        }

        *normals = (PyObject *)n;

}

void C_PyStaticRender_GenerateNormals(int mode, PyObject *vectors, PyObject **normals, PyObject **faces)
{
	switch (mode) {
		case GL_QUAD_STRIP:
			C_PyStaticRender_QuadStrip_GenNormals(vectors,
					normals, faces);
			break;
		default:
			printf("TODO: Build normals generator for mode %d!\n",
					mode);
	}
}

