/*
 * Copyright 2003, Shane Stephens & Jan Schmidt.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"
#include "PyVectorList_common.h"

int C_PyVectorList_Length(PyObject *v, int *result) {
        *result = ((PyVectorList *)v)->vectorsSize;
	return 1;
}

int C_PyVectorList_Values(PyObject *v, double **result) {
        *result = ((PyVectorList *)v)->values;
	return 1;
}

#if 0
static PyObject *C_PyVectorList_Clone(PyObject *self) {

        PyVectorList *ovl = (PyVectorList *)self;
        PyVectorList *nvl 
                = (PyVectorList *)C_PyVectorList_New();
        int i;
        double *p = nvl->values;
        PyVector **q = (PyVector **)nvl->vectors;

        nvl->vectorsSize = ovl->vectorsSize;
        
        memcpy(nvl->values, ovl->values, 3 * sizeof(double) * ovl->vectorsSize);

        for (i = 0; i < ovl->vectorsSize; i++)
        {
                (*q++) = (PyVector *)C_PyVector_New(p[0], p[1], p[2]);
                p += 3;
        }

        return (PyObject *)nvl;

}
        

static PyObject *PyVectorList_Clone(PyObject *self, PyObject *args) {

        return C_PyVectorList_Clone(self);

}

static PyObject *C_PyVectorList_Reverse(PyObject *self)
{

        PyVectorList *vl = (PyVectorList *)self;
        PyVectorList *nvl = 
                (PyVectorList *)C_PyVectorList_New();
        int i;

        for (i = 0; i < vl->vectorsSize; i++) {

                memcpy(nvl->values + 3 * i, 
                       vl->values + 3 * (vl->vectorsSize - i - 1),
                       3 * sizeof(double));
                
        }

        nvl->vectorsSize = vl->vectorsSize;

        for(i = 0; i < nvl->vectorsSize; i++) {
                nvl->vectors[i] = C_PyVector_New(nvl->values[3 * i],
				nvl->values[3 * i + 1], nvl->values[3 * i + 2]);
        }

        return (PyObject *)nvl;
        
        
}

static PyObject *PyVectorList_Reverse(PyObject *self, PyObject *args) {

        return C_PyVectorList_Reverse(self);

}

static void C_PyVectorList_Transform(PyObject *self, double matrix[3][3],
                PyObject *vector) {

        double nx, ny, nz;
        int i;
        PyVectorList *vl = (PyVectorList *)self;
        PyVector *v = (PyVector *)vector;
        double *p = vl->values;
        
        for (i = 0; i < vl->vectorsSize; i++) {
                nx = p[0]*matrix[0][0] + p[1]*matrix[0][1] + p[2]*matrix[0][2];
                ny = p[0]*matrix[1][0] + p[1]*matrix[1][1] + p[2]*matrix[1][2];
                nz = p[0]*matrix[2][0] + p[1]*matrix[2][1] + p[2]*matrix[2][2];

                p[0] = nx;
                p[1] = ny;
                p[2] = nz;

                p += 3;

        }
        
        p = vl->values;
        
        for (i = 0; i < vl->vectorsSize; i++) {
                *p++ += v->vector[0];
                *p++ += v->vector[1];
                *p++ += v->vector[2];
        }
}

static PyObject *PyVectorList_Transform(PyObject *self, PyObject *args) {

        PyObject *m = NULL, *v = NULL;
        PyObject *f;
        double matrix[3][3];
        int i, j;

        if (!PyArg_ParseTuple(args, "OO", &m, &v)) {
                return NULL;
        }

        for(i = 0; i < 3; i++) { /* col */
                for (j = 0; j < 3; j++) { /* row */
                        f = PyList_GetItem(m, i);
                        f = PyList_GetItem(f, j);
                        matrix[i][j] = PyFloat_AsDouble(f);
                }
        }

        C_PyVectorList_Transform(self, matrix, v);
        
        Py_INCREF(Py_None);
        return Py_None;

}

static void C_PyVectorList_Insert(PyObject *self, int pos, PyObject *vector) {

        PyVectorList *vl = (PyVectorList *)self;
        PyVector *v = (PyVector *)vector;
        int to_copy, i;

        if (vl->vectorsAlloc <= vl->vectorsSize) {
                if (vl->vectorsAlloc == 0) {
                        vl->vectorsAlloc++;
                } else {
                        vl->vectorsAlloc *= 2;
                }
                vl->vectors 
                        = realloc(vl->vectors, vl->vectorsAlloc * sizeof(void *));
                vl->values 
                        = realloc(vl->values, vl->vectorsAlloc * 3 * sizeof(double));
        }

        to_copy = vl->vectorsSize - pos;

        for(i = to_copy; i > 0; i--) {
                memcpy(vl->vectors + pos + i, vl->vectors + pos + i - 1,
                                sizeof(void *));
                memcpy(vl->values + 3 * (pos + i), 
                                vl->values + 3 * (pos + i - 1), 
                                sizeof(double) * 3);
        }

/*        
        memcpy(vl->vectors + pos + 1, vl->pyVectors + pos, 
                        to_copy * sizeof(void *));
        memcpy(vl->values + 3 * (pos + 1), vl->values + 3 * pos, 
                        to_copy * sizeof(double) * 3);
*/
        
        memcpy(vl->values + 3 * pos, v->vector, 3 * sizeof(double));
        
        if (v->memory_internal == TRUE) {
                free(v->vector);
                v->memory_internal = FALSE;
                v->vector = vl->values + 3 * pos;
        }

        vl->vectors[pos] = (PyObject *)v;

        Py_INCREF(v);
        vl->vectorsSize++;
        
}

static PyObject *PyVectorList_Insert(PyObject *self, PyObject *args) {

        PyObject *v = NULL;
        int pos = 0;

        if (!PyArg_ParseTuple(args, "iO", &pos, &v)) {
                return NULL;
        }

        C_PyVectorList_Insert(self, pos, v);

        Py_INCREF(Py_None);
        return Py_None;
        
}
static PyObject *PyVectorList_Item(PyObject *self, int index) {
	
	return C_PyVectorList_Item(self, index);
	
}
static PyObject *C_PyVectorList_Zip(PyObject *a, PyObject *b, 
		int a_cnt, int a_skip, int b_cnt, int b_skip) {
        PyVectorList *vla = (PyVectorList *)a;
        PyVectorList *vlb = (PyVectorList *)b;
        PyVectorList *result;
        int a_index, b_index;

        result = (PyVectorList *)C_PyVectorList_New();

	a_index = 0;
	b_index = 0;
	while ((a_index < vla->vectorsSize) || (b_index < vlb->vectorsSize))
        {
		int last_index;
		
		last_index = (a_index + a_cnt);
		if (last_index > vla->vectorsSize)
			last_index = vla->vectorsSize;

		while (a_index < last_index)
		{
			/* Take a_cnt entries from vla */
                	PyVector *tmp = (PyVector *)C_PyVector_New(
                        	vla->values[a_index * 3], 
				vla->values[a_index * 3 + 1],
				vla->values[a_index * 3 + 2]);
                	IVECTORLIST_APPEND((PyObject *)result, (PyObject *)tmp);
			a_index++;
		}
		a_index += a_skip;

		/* Take b_cnt entries from vlb */
		last_index = (b_index + b_cnt);
		if (last_index > vlb->vectorsSize)
			last_index = vlb->vectorsSize;

		while (b_index < last_index)
		{
			/* Take b_cnt entries from vlb */
                	PyVector *tmp = (PyVector *)C_PyVector_New(
				vlb->values[b_index * 3],
				vlb->values[b_index * 3 + 1],
				vlb->values[b_index * 3 + 2]);
                	IVECTORLIST_APPEND((PyObject *)result, (PyObject *)tmp);
			b_index++;
		}
		b_index += b_skip;
        }

        return (PyObject *)result;
        
}

PyObject *PyVectorList_Zip(PyObject *self, PyObject *args)
{
        PyObject *a = NULL, *b = NULL;
	int a_cnt=1, b_cnt=1;
	int a_skip=0, b_skip=0;

        if (!PyArg_ParseTuple(args, "OO|iiii", &a, &b, &a_cnt, &a_skip,
			      &b_cnt, &b_skip)) {
                return NULL;
        }

        return C_PyVectorList_Zip(a, b, a_cnt, a_skip, b_cnt, b_skip);

}
#endif 

PyObject *PyVectorList_Item(PyObject *self, int index);

int C_PyVectorList_Item(PyObject *self, int index, PyObject **result) {
	TOGRA_RETURN(PyVectorList_Item(self, index));
}
