/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"
#include "math.h"

#include "PyContainer_common.h"

PyObject *TograUnitSphere(PyObject *self, PyObject *args) {

	int slices, segs;
	double slice_height;
	double seg_angle;
	int slice, seg;
	double cur_slice;
	double cur_r;
	double last_slice;
	double last_r;
	
	if (!PyArg_ParseTuple(args, "ii", &slices, &segs)) {
		return NULL;
	}
	
	seg_angle = 2 * M_PI / segs;
	slice_height = 2.0 / slices;

	PyObject *top_slice = C_PyVectorList_New();
	PyObject *bot_slice = C_PyVectorList_New();
	PyObject *container;

	IVECTORLIST_APPEND(top_slice, C_PyVector_New(0.0, 1.0, 0.0));
	IVECTORLIST_APPEND(bot_slice, C_PyVector_New(0.0, -1.0, 0.0));	
	
	cur_slice = 1 - slice_height;
	cur_r = sqrt(1 -  cur_slice * cur_slice);

        for (seg = 0; seg <= segs; seg++) {
		IVECTORLIST_APPEND(top_slice, C_PyVector_New(
			cur_r * sin(seg * seg_angle), 
			cur_slice, 
			cur_r * cos(seg * seg_angle)));
		IVECTORLIST_APPEND(bot_slice, C_PyVector_New(
			cur_r * sin((segs - seg) * seg_angle), 
			-cur_slice, 
			cur_r * cos((segs - seg) * seg_angle)));
	}		

        container = C_PyContainer_New(1);
        /*
	((PySphere *)container)->slices = slices;
        ((PySphere *)container)->segs = segs;
	*/
        C_PyContainer_Append(container,
                C_PyStaticRender_New(top_slice, GL_TRIANGLE_FAN, top_slice, 
                                NULL, NULL, 0, 0, 0));

	for (slice = 1; slice < slices - 1; slice++) {
		
		PyObject *sl_obj = C_PyVectorList_New();
		
		last_slice = cur_slice;
		last_r = cur_r;
		
		cur_slice -= slice_height;
		cur_r = sqrt(1 - cur_slice * cur_slice);
		
		for (seg = segs; seg >= 0; seg--) {
                        if (cur_slice >= 0) {
                                IVECTORLIST_APPEND(sl_obj, C_PyVector_New(
                                        cur_r * sin(seg * seg_angle),
                                        cur_slice,
                                        cur_r * cos(seg * seg_angle)));
                                IVECTORLIST_APPEND(sl_obj, C_PyVector_New(
                                        last_r * sin(seg * seg_angle),
                                        last_slice,
                                        last_r * cos(seg * seg_angle)));
                        } else {
                                IVECTORLIST_APPEND(sl_obj, C_PyVector_New(
                                        cur_r * sin(seg * seg_angle),
                                        cur_slice,
                                        cur_r * cos(seg * seg_angle)));
                                IVECTORLIST_APPEND(sl_obj, C_PyVector_New(
                                        last_r * sin(seg * seg_angle),
                                        last_slice,
                                        last_r * cos(seg * seg_angle)));
                        }
		}
			
		C_PyContainer_Append(container,
                        C_PyStaticRender_New(sl_obj, GL_QUAD_STRIP, sl_obj, 
                                NULL, NULL, 0, 0, 0));
	}

        
	C_PyContainer_Append(container, 
                C_PyStaticRender_New(bot_slice, GL_TRIANGLE_FAN, bot_slice, 
                                NULL, NULL, 0, 0, 0));

	
	return container;

	
}

PyObject *TograUnitBox(PyObject *self, PyObject *args) {
	
	PyObject *vl = C_PyVectorList_New();
	PyObject *nl = C_PyVectorList_New();
	
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,-0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(1,0,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,-0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(1,0,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(1,0,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(1,0,0));
	
	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,-0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(-1,0,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,-0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(-1,0,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(-1,0,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(-1,0,0));	

	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,-0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,0,1));
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,-0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,0,1));
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,0,1));	
	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,0,1));

	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,-0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,0,-1));
	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,-0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,0,-1));
	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,0,-1));
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,0,-1));	

	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,-0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,-1,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,-0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,-1,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,-0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,-1,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,-0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,-1,0));	

	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,1,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(-0.5,0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,1,0));
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,0.5,0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,1,0));	
	IVECTORLIST_APPEND(vl, C_PyVector_New(0.5,0.5,-0.5));
	IVECTORLIST_APPEND(nl, C_PyVector_New(0,1,0));

	return C_PyStaticRender_New(vl, GL_QUADS, nl, NULL, NULL, 0, 0, 0);
	
}


PyObject *TograUnitCylinder(PyObject *self, PyObject *args) {

        int i;
        PyObject *top;
        PyObject *topn;
        PyObject *bottom;
        PyObject *bottomn;
        PyObject *sides;
        PyObject *sidesn;
        PyObject *container;
	int slices;

	if (!PyArg_ParseTuple(args, "i", &slices)) {
		return NULL;
	}
	
	top = C_PyVectorList_New();
	topn = C_PyVectorList_New();
	bottom = C_PyVectorList_New();
	bottomn = C_PyVectorList_New();
	sides = C_PyVectorList_New();
	sidesn = C_PyVectorList_New();
	container = C_PyContainer_New(1);
	
        IVECTORLIST_APPEND(top, C_PyVector_New(0, 0.5, 0));
        IVECTORLIST_APPEND(bottom, C_PyVector_New(0, -0.5, 0));
        IVECTORLIST_APPEND(topn, C_PyVector_New(0, 1, 0));
        IVECTORLIST_APPEND(bottomn, C_PyVector_New(0, -1, 0));

        for(i = 0; i <= slices; i++) {
                IVECTORLIST_APPEND(top, C_PyVector_New(
                                0.5*sin(i*2*M_PI/slices), 0.5,
                                0.5*cos(i*2*M_PI/slices)));
                IVECTORLIST_APPEND(topn, C_PyVector_New(0, 1, 0));
                IVECTORLIST_APPEND(bottom, C_PyVector_New(
                                0.5*sin((slices - i)*2*M_PI/slices), -0.5,
                                0.5*cos((slices - i)*2*M_PI/slices)));
                IVECTORLIST_APPEND(bottomn, C_PyVector_New(0, -1, 0));
                IVECTORLIST_APPEND(sides, C_PyVector_New(
                                0.5*sin(i*2*M_PI/slices), 0.5,
                                0.5*cos(i*2*M_PI/slices)));
                IVECTORLIST_APPEND(sides, C_PyVector_New(
                                0.5*sin(i*2*M_PI/slices), -0.5,
                                0.5*cos(i*2*M_PI/slices)));
                IVECTORLIST_APPEND(sidesn, C_PyVector_New(
                                0.5*sin(i*2*M_PI/slices), 0,
                                0.5*cos(i*2*M_PI/slices)));
                IVECTORLIST_APPEND(sidesn, C_PyVector_New(
                                0.5*sin(i*2*M_PI/slices), 0,
                                0.5*cos(i*2*M_PI/slices)));
        }
        
        C_PyContainer_Append(container, 
                C_PyStaticRender_New(top, GL_TRIANGLE_FAN, topn, NULL, NULL, 
                                0, 0, 0));
        
        C_PyContainer_Append(container, 
                C_PyStaticRender_New(sides, GL_QUAD_STRIP, sidesn, NULL, NULL, 
                                0, 0, 0));

        C_PyContainer_Append(container, 
                C_PyStaticRender_New(bottom, GL_TRIANGLE_FAN, bottomn, NULL, 
                                NULL, 0, 0, 0));
        
        return container;

}

/*
PyObject *C_TograCircle(double cx, double cy, double r, int steps) {

        PyObject *circle = C_PyVectorList_New()
        int i;

        for (i = 0; i < steps; i++)
        {
                VL_APPEND(circle, C_PyVector_New(
                                        cx + r * sin(i * 2 * M_PI / steps),
                                        cy + r * cos(i * 2 * M_PI / steps),
                                        0
                        ));
        }

        return circle;

}

PyObject *TograCircle(PyObject *self, PyObject *args) {

        double cx, cy, r;
        int steps;

        if (!PyArg_ParseTuple(args, "dddi", &cx, &cy, &r, &steps)) {
                return NULL;
        }

        return C_TograCircle(cx, cy, r, steps);

}

PyObject *C_TograExtrude(PyObject *vector_list, PyObject *extrude_path) {

        PyVectorList *vl = (PyVectorList *)vector_list;
        PyVector *v = (PyVector *)extrude_path;
        int i;
        double *p;
        PyVectorList *object_data;
        PyVector **q;

        object_data 
               = (PyVectorList *)C_PyVectorList_New()


        p = object_data->values;
        
        for (i = 0; i < vl->length; i++) {
                *p++ = vl->values[i*3];
                *p++ = vl->values[i*3 + 1];
                *p++ = vl->values[i*3 + 2];
                *p++ = vl->values[i*3] + v->vector[0];
                *p++ = vl->values[i*3 + 1] + v->vector[1];
                *p++ = vl->values[i*3 + 2] + v->vector[2];
        }

        *p++ = vl->values[0];
        *p++ = vl->values[1];
        *p++ = vl->values[2];
        *p++ = vl->values[0] + v->vector[0];
        *p++ = vl->values[1] + v->vector[1];
        *p++ = vl->values[2] + v->vector[2];
        
        object_data->length = vl->length * 2 + 2;

        q = (PyVector **)object_data->pyVectors;
        p = object_data->values;
        for (i = 0; i < object_data->length; i++)
        {
                (*q++) = (PyVector *)C_PyVector_FromVector(p);
                p += 3;
        }

        return (PyObject *)object_data;
                        
}

PyObject *TograExtrude(PyObject *self, PyObject *args) {

        PyObject *vl, *v;
        
        if (!PyArg_ParseTuple(args, "OO", &vl, &v)) {
                return NULL;
        }

        return C_TograExtrude(vl, v);
        
}
*/
