from Togra import *
from TograUtil import *
import math

class Tracker:
	def __init__(t, width, height, scene, tCont, min_period=100):
		t.width = width
		t.height = height
		t.diff = 0
		t.scene = scene
		t.tCont = tCont
		t.min_period=min_period
		t.is_pressed = False
                t.start = None
		
	def calc_track_vector(t, x, y):
		v = [0,0,0]
		v[0] = (2.0 * x - t.width) / t.width
		v[1] = -(t.height - 2.0 * y) / t.height
		d = math.sqrt(v[0]*v[0] + v[1]*v[1])
		#if d > 1.0:
		#	d = 1.0
		v[2] = math.cos((3.14159/2) * d)
		a = 1.0 / math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])
		v[0] *= a
		v[1] *= a
		v[2] *= a
		return v

	def mouseFunc(t, args):
		t.is_pressed = args[1]
		if t.is_pressed:
			t.time = t.scene.getTime()
			if (t.diff > 0):
				bt = t.tCont[1].getBaseTime()
				diff2 = (t.time - bt) % t.diff
				angle = diff2 / t.diff * 360
				rMat = Rotation(
					Vector(t.axis[0], t.axis[1], t.axis[2]),
					int(angle))

				m = rMat.multiply(t.tCont[0])
				t.tCont[0] = m
			t.tCont[1] = Translation(Vector(0,0,0))

			t.start = t.calc_track_vector(args[2], args[3])
		else:
			diff = t.scene.getTime() - t.time
			if (diff > 0):
				t.diff = 0
				return
			t.diff = t.olddiff * 360/t.angle
			if t.diff < t.min_period:
				t.diff = t.min_period
			t.tCont[1] = rotatePath(
				Vector(t.axis[0], t.axis[1], t.axis[2]), 
				int(t.diff))

	def motionFunc(t, args):
		if t.is_pressed and t.start:
			end = t.calc_track_vector(args[0], args[1])
			dx = end[0] - t.start[0]
			dy = end[1] - t.start[1]
			dz = end[2] - t.start[2]
			t.angle = 90.0 * \
				math.sqrt((dx * dx + dy * dy + dz * dz) / 2)
			axis = [0,0,0]
			axis[0] = t.start[1] * end[2] - t.start[2] * end[1]
			axis[1] = t.start[2] * end[0] - t.start[0] * end[2]
			axis[2] = t.start[0] * end[1] - t.start[1] * end[0]
			t.axis = axis
			t.start = end
			rMat = Rotation(Vector(axis[0], axis[1], axis[2]), 
				t.angle)
			m = rMat.multiply(t.tCont[0])
			t.tCont[0] = m
			newtime = t.scene.getTime()
			t.olddiff = newtime - t.time
			t.time = newtime

def initTracker(scene):
	pauseQuitKeystrokes(scene)
	t = TransformContainer()
	t.append(Translation(Vector(0,0,0)))
	t.append(Translation(Vector(0,0,0)))

	tr = Tracker(800, 800, scene, t)
	def mouseFunc(*args):
		tr.mouseFunc(args)
	def motionFunc(*args):
		tr.motionFunc(args)
	scene.mouseFunction = mouseFunc
	scene.motionFunction = motionFunc
	return t


